;==============================================================================

; file: "_kernel.scm"

; Copyright (C) 1994, 1995 by Marc Feeley, All Rights Reserved.

(##include "header.scm")

(declare
  (multilisp)
  (standard-bindings)
  (extended-bindings)
  (block)
  (fixnum)
)

(declare (not interrupts-enabled))

(##c-declaration "

/*---------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

#include \"setup.h\"
#include \"os.h\"

___DEMAND_GLO(___G__23_23exception_2estack_2doverflow)
___DEMAND_GLO(___G__23_23exception_2eheap_2doverflow)
___DEMAND_GLO(___G__23_23exception_2euser_2dinterrupt)
___DEMAND_GLO(___G__23_23exception_2etimer_2dinterrupt)
___DEMAND_GLO(___G__23_23exception_2ewrong_2dnb_2darg)
___DEMAND_GLO(___G__23_23exception_2ec_2dconv_2derror)

extern long semispace_length;
extern long semispace_fudge;
extern long stack_length;
extern long stack_fudge;

___WORD ___internal_return;

#ifdef sparc
#define CC_PROGRAM \"cc\"
#include <dlfcn.h>
#ifdef RTLD_NOW
#define LD_PROGRAM \"/usr/ccs/bin/ld -G\"
#else
#define LD_PROGRAM \"/usr/bin/ld\"
#endif
#endif

#ifdef hpux
#define CC_PROGRAM \"cc +z\"
#define LD_PROGRAM \"/bin/ld -b\"
#endif

#ifdef sgi
#define CC_PROGRAM \"cc\"
#define LD_PROGRAM \"/usr/bin/ld -shared\"
#endif

#ifdef __alpha
#define CC_PROGRAM \"cc -shared\"
#define LD_PROGRAM \"/usr/bin/ld -shared -expect_unresolved __exc_\\*\"
#endif

#ifndef CC_PROGRAM
#define CC_PROGRAM NULL
#endif

#ifndef LD_PROGRAM
#define LD_PROGRAM NULL
#endif

/*---------------------------------------------------------------------------*/

")

(define (##setup-traps)

  (##c-code "
    ___ps->handler_break           = ___LBL(1);
    ___ps->handler_stack_limit     = ___LBL(2);
    ___ps->handler_heap_limit      = ___LBL(3);
    ___ps->handler_wrong_nargs     = ___LBL(4);
    ___ps->handler_wrong_nargs_clo = ___LBL(5);
    ___ps->handler_build_rest      = ___LBL(6);
    ___ps->handler_build_rest_clo  = ___LBL(7);
    ___ps->handler_touch           = ___LBL(8);
    ___ps->handler_c_conv_error    = ___LBL(9);
    ___internal_return             = ___LBL(10);
    os_init();
    ___sp[0] = ___ps->handler_break;
    ___JUMPEXT(___NOTHING,___R0)
    ")

  (##c-code "
    /* This code is at ___LBL(1) */

{
  ___WORD ra    = ___sp[0]; /* return address of parent continuation  */
  ___WORD frame = ___sp[1]; /* pointer to parent continuation's frame */

/*
  if (___ps->sb != ___sp)
    { printf(\"*** ERROR -- sb != sp\\n\"); exit(1); }
*/

  if (___TYP(frame)==___tSUBTYPED)
    {
      /* continuation's frame is in the heap */

      ___WORD *fp;
      int fs, link, i;

      fp = (___WORD*)(frame-___tSUBTYPED+___WS);

      if (ra == ___internal_return)
        {
          ___WORD fs_link = ((___WORD *)(fp[1]-___tPROCEDURE))[0];
          fs = (fs_link>>3)&((1<<14)-1);
          link = (fs_link>>17)&((1<<14)-1);
        }
      else
        {
          ___WORD fs_link = ((___WORD *)(ra-___tPROCEDURE))[0];
          fs = (fs_link>>3)&((1<<14)-1);
          link = (fs_link>>17)&((1<<14)-1);
        }
      fp += fs+1;

      ___sp[0] = fp[-link-1];
      ___sp[1] = fp[-fs-1];

      for (i=-fs; i<0; i++)
        ___sp[i] = fp[i];

      ___sp[-link-1] = ___ps->handler_break;
      ___sp -= fs;

      ___JUMPEXT(___NOTHING,ra)
    }
  else
    {
      /* continuation's frame is in the stack */

      ___WORD *fp, frame_ra;
      int fs, link, i;

      if (ra == ___internal_return)
        {
          ___WORD fs_link = ((___WORD *)(((___WORD*)frame)[0]-___tPROCEDURE))[0];
          fs = (fs_link>>3)&((1<<14)-1);
          link = (fs_link>>17)&((1<<14)-1);
        }
      else
        {
          ___WORD fs_link = ((___WORD *)(ra-___tPROCEDURE))[0];
          fs = (fs_link>>3)&((1<<14)-1);
          link = (fs_link>>17)&((1<<14)-1);
        }
      fp = ((___WORD*)frame)+fs;
      frame_ra = fp[-link-1];

      if (frame_ra == ___ps->handler_break)
        {
          /* first frame of that section */

          ___sp[0] = fp[0];
          ___sp[1] = fp[1];

          for (i=-fs; i<0; i++)
            ___sp[i] = fp[i];

          ___sp -= fs;

          ___JUMPEXT(___NOTHING,ra)
        }
      else
        {
          /* not the first frame of that section */

          ___sp[0] = frame_ra;
          ___sp[1] = (___WORD)fp;

          for (i=-fs; i<0; i++)
            ___sp[i] = fp[i];

          ___sp[-link-1] = ___ps->handler_break;
          ___sp -= fs;

          ___JUMPEXT(___NOTHING,ra)
        }
    }
}" (0))

  (##c-code "
    /* This code is at ___LBL(2) */

    {
      ___WORD *sl = ___ps->sl;
      if (sl == ___ps->stack_base)
      {
        if (___ps->intr_enabled)
        {
          ___ps->sl = ___ps->stack+stack_fudge;
          if (___ps->intr_flag[0])
          {
            ___ps->intr_flag[0] = 0;
            ___ps->sl = ___ps->stack+stack_fudge;
            ___SET_PUSH(1,___R0)
            ___SET_PUSH(2,___R1)
            ___SET_PUSH(3,___R2)
            ___SET_PUSH(4,___R3)
            ___SET_PUSH(5,___R4)
            ___SET_PUSH(6,___ps->temp1)
            ___ADJFP(6)
            ___SET_R0(___internal_return)
            ___JUMPEXT(___SET_NARGS(0),___G__23_23exception_2euser_2dinterrupt.value)
          }
          else if (___ps->intr_flag[1])
          {
            ___ps->intr_flag[1] = 0;
            ___ps->sl = ___ps->stack+stack_fudge;
            ___SET_PUSH(1,___R0)
            ___SET_PUSH(2,___R1)
            ___SET_PUSH(3,___R2)
            ___SET_PUSH(4,___R3)
            ___SET_PUSH(5,___R4)
            ___SET_PUSH(6,___ps->temp1)
            ___ADJFP(6)
            ___SET_R0(___internal_return)
            ___JUMPEXT(___SET_NARGS(0),___G__23_23exception_2etimer_2dinterrupt.value)
          }
        }
        ___JUMPEXT(___NOTHING,___ps->temp1)
      }
      else
      {
        ___W_ALL gc( ___pstate, 0 ); ___R_ALL
        if (___sp<___ps->stack+stack_fudge)
        {
          printf(\"*** ERROR -- Fatal stack overflow\\n\");
          exit(1);
        }
        ___JUMPEXT(___NOTHING,___ps->temp1)
/*
        ___SET_PUSH(1,___R0)
        ___SET_PUSH(2,___R1)
        ___SET_PUSH(3,___R2)
        ___SET_PUSH(4,___R3)
        ___SET_PUSH(5,___R4)
        ___SET_PUSH(6,___ps->temp1)
        ___ADJFP(6)
        ___SET_R0(___internal_return)
        ___JUMPEXT(___SET_NARGS(0),___G__23_23exception_2estack_2doverflow.value)
*/
      }
    }" (0))

  (##c-code "
    /* This code is at ___LBL(3) */
    ___W_ALL gc( ___pstate, 1024*___WS ); ___R_ALL
    ___JUMPEXT(___NOTHING,___ps->temp1)
    " (0))

  (##c-code "
    /* This code is at ___LBL(4) */

    {
      int i, nb_args = ___ps->na+1;

      switch (___ps->na)
      {
        case 0:  /* */
          break;
        case 1:  /* ___R1=arg1 */
          ___SET_PUSH(1,___R1)
          ___ADJFP(1)
          break;
        case 2:  /* ___R1=arg1 ___R2=arg2 */
          ___SET_PUSH(1,___R1)
          ___SET_PUSH(2,___R2)
          ___ADJFP(2)
          break;
        default: /* ___R1=argN-2 ___R2=argN-1 ___R3=argN */
          ___SET_PUSH(1,___R1)
          ___SET_PUSH(2,___R2)
          ___SET_PUSH(3,___R3)
          ___ADJFP(3)
      }

      ___SET_PUSH(1,0)
      ___ADJFP(1)

      for (i=1; i<nb_args; i++)
        ___sp[i-1] = ___sp[i];
      ___sp[nb_args-1] = ___ps->temp1;

      switch (nb_args)
      {
        case 1:
          ___SET_R1(___POP(0))
          ___ADJFP(-1)
          break;
        case 2:
          ___SET_R2(___POP(0))
          ___SET_R1(___POP(1))
          ___ADJFP(-2)
          break;
        default:
          ___SET_R3(___POP(0))
          ___SET_R2(___POP(1))
          ___SET_R1(___POP(2))
          ___ADJFP(-3)
      }

      ___JUMPEXT(___SET_NARGS(nb_args),___G__23_23exception_2ewrong_2dnb_2darg.value)
    }
    " (0))

  (##c-code "
    /* This code is at ___LBL(5) */

    {
      int i, nb_args = ___ps->na+1;

      switch (___ps->na)
      {
        case 0:  /* */
          break;
        case 1:  /* ___R1=arg1 */
          ___SET_PUSH(1,___R1)
          ___ADJFP(1)
          break;
        case 2:  /* ___R1=arg1 ___R2=arg2 */
          ___SET_PUSH(1,___R1)
          ___SET_PUSH(2,___R2)
          ___ADJFP(2)
          break;
        default: /* ___R1=argN-2 ___R2=argN-1 ___R3=argN */
          ___SET_PUSH(1,___R1)
          ___SET_PUSH(2,___R2)
          ___SET_PUSH(3,___R3)
          ___ADJFP(3)
      }

      ___SET_PUSH(1,0)
      ___ADJFP(1)

      for (i=1; i<nb_args; i++)
        ___sp[i-1] = ___sp[i];
      ___sp[nb_args-1] = ___r4;

      switch (nb_args)
      {
        case 1:
          ___SET_R1(___POP(0))
          ___ADJFP(-1)
          break;
        case 2:
          ___SET_R2(___POP(0))
          ___SET_R1(___POP(1))
          ___ADJFP(-2)
          break;
        default:
          ___SET_R3(___POP(0))
          ___SET_R2(___POP(1))
          ___SET_R1(___POP(2))
          ___ADJFP(-3)
      }

      ___JUMPEXT(___SET_NARGS(nb_args),___G__23_23exception_2ewrong_2dnb_2darg.value)
    }
    " (0))

  (##c-code "
    /* This code is at ___LBL(6) */

    {
      int n;
      ___WORD z = ((___WORD *)(___ps->temp1-___tPROCEDURE))[0];
      int nb_parms = (z>>3)&((1<<14)-1);

      switch (___ps->na)
      {
        case 0:  /* */
          break;
        case 1:  /* ___R1=arg1 */
          ___SET_PUSH(1,___R1)
          ___ADJFP(1)
          break;
        case 2:  /* ___R1=arg1 ___R2=arg2 */
          ___SET_PUSH(1,___R1)
          ___SET_PUSH(2,___R2)
          ___ADJFP(2)
          break;
        default: /* ___R1=argN-2 ___R2=argN-1 ___R3=argN */
          ___SET_PUSH(1,___R1)
          ___SET_PUSH(2,___R2)
          ___SET_PUSH(3,___R3)
          ___ADJFP(3)
      }

      if (___ps->na<nb_parms)
      {
      int i, nb_args = ___ps->na+1;
      ___SET_PUSH(1,0)
      ___ADJFP(1)

      for (i=1; i<nb_args; i++)
        ___sp[i-1] = ___sp[i];
      ___sp[nb_args-1] = ___ps->temp1;

      switch (nb_args)
      {
        case 1:
          ___SET_R1(___POP(0))
          ___ADJFP(-1)
          break;
        case 2:
          ___SET_R2(___POP(0))
          ___SET_R1(___POP(1))
          ___ADJFP(-2)
          break;
        default:
          ___SET_R3(___POP(0))
          ___SET_R2(___POP(1))
          ___SET_R1(___POP(2))
          ___ADJFP(-3)
      }

      ___JUMPEXT(___SET_NARGS(nb_args),___G__23_23exception_2ewrong_2dnb_2darg.value)
      }

      ___SET_R1(___NUL)
      n = ___ps->na - nb_parms + 1;
      while (n > 0)
      {
        ___SET_R1(___CONS(___POP(0),___R1))
        ___ADJFP(-1)
        n--;
      }

      switch (nb_parms)
      {
        case 1:  /* ___R1=rest */
          break;
        case 2:  /* ___R1=parm1 ___R2=rest */
          ___SET_R2(___R1)
          ___SET_R1(___POP(0))
          ___ADJFP(-1)
          break;
        default: /* ___R1=parmN-2 ___R2=parmN-1 ___R3=rest */
          ___SET_R3(___R1)
          ___SET_R2(___POP(0))
          ___SET_R1(___POP(1))
          ___ADJFP(-2)
      }

      ___JUMPEXT(___SET_NARGS(-1),___ps->temp1)
    }
    " (0))

  (##c-code "
    /* This code is at ___LBL(7) */

    {
      int n;
      ___WORD p = ((___WORD *)(___R4-___tPROCEDURE))[0];
      ___WORD z = ((___WORD *)(p-___tPROCEDURE))[0];
      int nb_parms = (z>>3)&((1<<14)-1);

      switch (___ps->na)
      {
        case 0:  /* */
          break;
        case 1:  /* ___R1=arg1 */
          ___SET_PUSH(1,___R1)
          ___ADJFP(1)
          break;
        case 2:  /* ___R1=arg1 ___R2=arg2 */
          ___SET_PUSH(1,___R1)
          ___SET_PUSH(2,___R2)
          ___ADJFP(2)
          break;
        default: /* ___R1=argN-2 ___R2=argN-1 ___R3=argN */
          ___SET_PUSH(1,___R1)
          ___SET_PUSH(2,___R2)
          ___SET_PUSH(3,___R3)
          ___ADJFP(3)
      }

      if (___ps->na<nb_parms)
      {
      int i, nb_args = ___ps->na+1;
      ___SET_PUSH(1,0)
      ___ADJFP(1)

      for (i=1; i<nb_args; i++)
        ___sp[i-1] = ___sp[i];
      ___sp[nb_args-1] = ___ps->temp1;

      switch (nb_args)
      {
        case 1:
          ___SET_R1(___POP(0))
          ___ADJFP(-1)
          break;
        case 2:
          ___SET_R2(___POP(0))
          ___SET_R1(___POP(1))
          ___ADJFP(-2)
          break;
        default:
          ___SET_R3(___POP(0))
          ___SET_R2(___POP(1))
          ___SET_R1(___POP(2))
          ___ADJFP(-3)
      }

      ___JUMPEXT(___SET_NARGS(nb_args),___G__23_23exception_2ewrong_2dnb_2darg.value)
      }

      ___SET_R1(___NUL)
      n = ___ps->na - nb_parms + 1;
      while (n > 0)
      {
        ___SET_R1(___CONS(___POP(0),___R1))
        ___ADJFP(-1)
        n--;
      }

      switch (nb_parms)
      {
        case 1:  /* ___R1=rest */
          break;
        case 2:  /* ___R1=parm1 ___R2=rest */
          ___SET_R2(___R1)
          ___SET_R1(___POP(0))
          ___ADJFP(-1)
          break;
        default: /* ___R1=parmN-2 ___R2=parmN-1 ___R3=rest */
          ___SET_R3(___R1)
          ___SET_R2(___POP(0))
          ___SET_R1(___POP(1))
          ___ADJFP(-2)
      }

      ___JUMPEXT(___SET_NARGS(-1),___R4)
    }
    " (0))

  (##c-code "
    /* This code is at ___LBL(8) */

    {
      ___WORD x = ___ps->temp2;
      ___WORD y = ___UNBOX(x,___tPLACEHOLDER)[0];
      if (x != y)
      {
        ___ps->temp2 = y;
        ___JUMPEXT(___NOTHING,___ps->temp1)
      }
      else
      {
        ___WORD z = ___UNBOX(x,___tPLACEHOLDER)[1];
        ___SET_PUSH(1,___R0)
        ___SET_PUSH(2,___R1)
        ___SET_PUSH(3,___R2)
        ___SET_PUSH(4,___R3)
        ___SET_PUSH(5,___R4)
        ___SET_PUSH(6,___ps->temp1)
        ___ADJFP(6)
        ___SET_R1(x)
        ___SET_R2(z)
        ___SET_R0(___internal_return)
        ___JUMPEXT(___SET_NARGS(2),___G__23_23touch_2dundetermined.value)
      }
    }
    " (0))

  (##c-code "
    /* This code is at ___LBL(9) */

    {
      int i, nb_args = ___ps->na+1;

      switch (___ps->na)
      {
        case 0:  /* */
          break;
        case 1:  /* ___R1=arg1 */
          ___SET_PUSH(1,___R1)
          ___ADJFP(1)
          break;
        case 2:  /* ___R1=arg1 ___R2=arg2 */
          ___SET_PUSH(1,___R1)
          ___SET_PUSH(2,___R2)
          ___ADJFP(2)
          break;
        default: /* ___R1=argN-2 ___R2=argN-1 ___R3=argN */
          ___SET_PUSH(1,___R1)
          ___SET_PUSH(2,___R2)
          ___SET_PUSH(3,___R3)
          ___ADJFP(3)
      }

      ___SET_PUSH(1,0)
      ___ADJFP(1)

      for (i=1; i<nb_args; i++)
        ___sp[i-1] = ___sp[i];
      ___sp[nb_args-1] = ___ps->temp1;

      switch (nb_args)
      {
        case 1:
          ___SET_R1(___POP(0))
          ___ADJFP(-1)
          break;
        case 2:
          ___SET_R2(___POP(0))
          ___SET_R1(___POP(1))
          ___ADJFP(-2)
          break;
        default:
          ___SET_R3(___POP(0))
          ___SET_R2(___POP(1))
          ___SET_R1(___POP(2))
          ___ADJFP(-3)
      }

      ___JUMPEXT(___SET_NARGS(nb_args),___G__23_23exception_2ec_2dconv_2derror.value)
    }
    " (0))

  (##c-code "
    /* This code is at ___LBL(10) */
    ___ps->temp2 = ___R1; /* for return from ##touch-undetermined */
    ___temp = ___POP(6);
    ___SET_R4(___POP(5))
    ___SET_R3(___POP(4))
    ___SET_R2(___POP(3))
    ___SET_R1(___POP(2))
    ___SET_R0(___POP(1))
    ___ADJFP(-6)
    ___JUMPEXT(___NOTHING,___temp)
    " (0))

)

(##setup-traps)

(declare (interrupts-enabled))

;------------------------------------------------------------------------------

; Memory allocation

(define (##make-vector n val)
  (let ((v (##c-code "
{
long words = ((___INT(___ARG1)*___WS+___WS+(1<<___TB)-1)&-(1<<___TB))/___WS;
if (___hp-___ps->hl<words)
{
___W_ALL gc( ___pstate, words ); ___R_ALL
}
___hp -= words;
___RESULT = (___WORD)___hp + ___tSUBTYPED;
___HEADER(___RESULT)=(___INT(___ARG1)<<___SB+___TB)+(___sVECTOR<<___TB);
}
" n)))
    (let loop ((i (##fixnum.- n 1)))
      (if (##fixnum.< i 0)
        v
        (begin
          (##vector-set! v i val)
          (loop (##fixnum.- i 1)))))))

(define (##make-string n val)
  (let ((s (##c-code "
{
long words = ((___INT(___ARG1)+1+___WS+(1<<___TB)-1)&-(1<<___TB))/___WS;
if (___hp-___ps->hl<words)
{
___W_ALL gc( ___pstate, words ); ___R_ALL
}
___hp -= words;
___RESULT = (___WORD)___hp + ___tSUBTYPED;
___HEADER(___RESULT)=(___INT(___ARG1)+1<<___SB+___TB)+(___sSTRING<<___TB);
}
" n)))
    (##string-set! s n #\nul)
    (let loop ((i (##fixnum.- n 1)))
      (if (##fixnum.< i 0)
        s
        (begin
          (##string-set! s i val)
          (loop (##fixnum.- i 1)))))))

(define (##make-vector8 n val)
  (let ((v (##c-code "
{
long words = ((___INT(___ARG1)+___WS+(1<<___TB)-1)&-(1<<___TB))/___WS;
if (___hp-___ps->hl<words)
{
___W_ALL gc( ___pstate, words ); ___R_ALL
}
___hp -= words;
___RESULT = (___WORD)___hp + ___tSUBTYPED;
___HEADER(___RESULT)=(___INT(___ARG1)<<___SB+___TB)+(___sSTRING<<___TB);
}
" n)))
    (let loop ((i (##fixnum.- n 1)))
      (if (##fixnum.< i 0)
        v
        (begin
          (##vector8-set! v i val)
          (loop (##fixnum.- i 1)))))))

(define (##make-vector16 n val)
  (let ((v (##c-code "
{
long words = (((___INT(___ARG1)*2)+___WS+(1<<___TB)-1)&-(1<<___TB))/___WS;
if (___hp-___ps->hl<words)
{
___W_ALL gc( ___pstate, words ); ___R_ALL
}
___hp -= words;
___RESULT = (___WORD)___hp + ___tSUBTYPED;
___HEADER(___RESULT)=((___INT(___ARG1)*2)<<___SB+___TB)+(___sSTRING<<___TB);
}
" n)))
    (let loop ((i (##fixnum.- n 1)))
      (if (##fixnum.< i 0)
        v
        (begin
          (##vector16-set! v i val)
          (loop (##fixnum.- i 1)))))))

;------------------------------------------------------------------------------

; I/O

(define (##display-string str)
  (let ((n (##string-length str)))
    (let loop ((i 0))
      (if (##fixnum.< i n)
        (begin
          (##c-code "putchar(___INT(___ARG1));___RESULT=___FAL;"
                    (##string-ref str i))
          (loop (##fixnum.+ i 1)))))))

(define (##err msg)
  (##display-string "*** ERROR -- Undefined procedure: ")
  (##display-string msg)
  (##display-string "
")
  (##c-code "exit(1);"))


(define stdio#stdin  (##c-procedure () pointer "___result = stdin;"))
(define stdio#stdout (##c-procedure () pointer "___result = stdout;"))
(define stdio#stderr (##c-procedure () pointer "___result = stderr;"))

(define stdio#fopen  (##c-procedure (string string) pointer "fopen"))
(define stdio#fclose (##c-procedure (pointer) int "fclose"))
(define stdio#fread  (##c-procedure (pointer long long pointer) long "fread"))
(define stdio#fwrite (##c-procedure (pointer long long pointer) long "fwrite"))

(define (##os-file-open-input filename)        (stdio#fopen filename "r"))
(define (##os-file-open-input-output filename) (stdio#fopen filename "r+"))
(define (##os-file-open-output filename)       (stdio#fopen filename "w"))

(define (##os-file-close descr)
  (##fixnum.= (stdio#fclose descr) 0))

(define (##os-file-read descr s i j)
  (##c-code
    "{
       FILE *stream = (___ARG1==___FIX(0)) ? stdin :
                      (___ARG1==___FIX(1)) ? stdout :
                      (___ARG1==___FIX(2)) ? stderr :
                      *(FILE **)(___ARG1-___tSUBTYPED+___WS);
       int n = fread( (char *)(___ARG2-___tSUBTYPED+___WS+___INT(___ARG3)),
                      sizeof( char ),
                      ___INT(___ARG4)-___INT(___ARG3),
                      stream );
       if (feof( stream )) clearerr( stream );
#ifndef __MWERKS__
       else if ((n==0)&&(errno==EINTR)) { n = -1; errno=0; }
#endif
       ___RESULT = ___FIX(n);
     }"
    descr
    s
    i
    j))

(define (##os-file-read-ready descr)
  #t)

(define (##os-file-write descr s i j)
  (##c-code
    "{
       FILE *stream = (___ARG1==___FIX(0)) ? stdin :
                      (___ARG1==___FIX(1)) ? stdout :
                      (___ARG1==___FIX(2)) ? stderr :
                      *(FILE **)(___ARG1-___tSUBTYPED+___WS);
       int n = fwrite( (char *)(___ARG2-___tSUBTYPED+___WS+___INT(___ARG3)),
                       sizeof( char ),
                       ___INT(___ARG4)-___INT(___ARG3),
                       stream );
       ___RESULT = ___FIX(n);
     }"
    descr
    s
    i
    j))

;------------------------------------------------------------------------------

; Procedure application and continuations

(define (##apply proc args)
  (##c-code "
{
  ___WORD proc = ___ARG1; /* procedure to call   */
  ___WORD args = ___ARG2; /* arguments to pass   */
  int n = 0;              /* number of arguments */

  while (args != ___NUL)
    {
      ___SET_PUSH(1,___CAR(args))
      ___ADJFP(1)
      args = ___CDR(args);
      n++;
    }

  switch (n)
    {
      case 0:  /* no arguments */
        break;
#if ___NB_ARG_REGS > 1
      case 1:  /* R1=arg1 */
#else
      default: /* R1=argN */
#endif
        ___SET_R1(___POP(0))
        ___ADJFP(-1)
#if ___NB_ARG_REGS > 1
        break;
#if ___NB_ARG_REGS > 2
      case 2:  /* R1=arg1 R2=arg2 */
#else
      default: /* R1=argN-1 R2=argN */
#endif
        ___SET_R2(___POP(0))
        ___SET_R1(___POP(1))
        ___ADJFP(-2)
#if ___NB_ARG_REGS > 2
        break;
#if ___NB_ARG_REGS > 3
      case 3:  /* R1=arg1 R2=arg2 R3=arg3 */
#else
      default: /* R1=argN-2 R2=argN-1 R3=argN */
#endif
        ___SET_R3(___POP(0))
        ___SET_R2(___POP(1))
        ___SET_R1(___POP(2))
        ___ADJFP(-3)
#if ___NB_ARG_REGS > 3
        break;
#if ___NB_ARG_REGS > 4
      case 4:  /* R1=arg1 R2=arg2 R3=arg3 R4=arg4 */
#else
      default: /* R1=argN-3 R2=argN-2 R3=argN-1 R4=argN */
#endif
        ___SET_R4(___POP(0))
        ___SET_R3(___POP(1))
        ___SET_R2(___POP(2))
        ___SET_R1(___POP(3))
        ___ADJFP(-4)
#if ___NB_ARG_REGS > 4
        break;
#if ___NB_ARG_REGS > 5
      case 5:  /* R1=arg1 R2=arg2 R3=arg3 R4=arg4 R5=arg5 */
#else
      default: /* R1=argN-4 R2=argN-3 R3=argN-2 R4=argN-1 R5=argN */
#endif
        ___SET_R5(___POP(0))
        ___SET_R4(___POP(1))
        ___SET_R3(___POP(2))
        ___SET_R2(___POP(3))
        ___SET_R1(___POP(4))
        ___ADJFP(-5)
#if ___NB_ARG_REGS > 5
        break;
      default: /* R1=argN-5 R2=argN-4 R3=argN-3 R4=argN-2 R5=argN-1 R6=argN */
        ___SET_R6(___POP(0))
        ___SET_R5(___POP(1))
        ___SET_R4(___POP(2))
        ___SET_R3(___POP(3))
        ___SET_R2(___POP(4))
        ___SET_R1(___POP(5))
        ___ADJFP(-6)
#endif
#endif
#endif
#endif
#endif
    }

  ___JUMPEXT(___SET_NARGS(n),proc)
}"
    proc
    args))

(define (##call-with-current-continuation p)
  (let ((v (##c-code "
{
  ___WORD frame, ra;
  if (___R0 == ___ps->handler_break)
    {
      frame = ___sp[1];
      ra    = ___sp[0];
    }
  else
    {
      frame = (___WORD)___sp;
      ra    = ___R0;
      ___sp[-1] = frame;
      ___sp[-2] = ra;
      ___sp -= 2;
      ___ps->sb = ___sp;
      ___R0 = ___ps->handler_break;
    }

  ___hp-=((((3)+1)*___WS+___AU-1)&-___AU)/___WS;
  {___WORD *___ptr=___hp+1;
  *___hp=((3)<<___SB+___TB)+(___sCONTFRAME<<___TB);
  ___STORE(frame)
  ___STORE(ra)
  ___STORE(___ARG1)}
  ___RESULT = ___GET_VECTOR;
}" ##dynamic-environment)))
    (p
      (lambda (x)
        (##c-code "
{
  ___WORD v = ___ARG1;
  ___WORD x = ___ARG2;
  ___sp = ___ps->sb;
  ___sp[1] = ___VECTORREF(v,___FIX(0));
  ___sp[0] = ___VECTORREF(v,___FIX(1));
  ___G__23_23dynamic_2denvironment.value = ___VECTORREF(v,___FIX(2));
  ___SET_R1(x)
  ___JUMPEXT(___NOTHING,___ps->handler_break)
}" v x)))))

(define (##continuation->frame c)
  (##closure-ref c 1))

(define (##frame-ret f)
  (##c-code "
{
  ___WORD f = ___ARG1;
  ___WORD frame = ___VECTORREF(f,___FIX(0));
  ___WORD ra    = ___VECTORREF(f,___FIX(1));

  if (___TYP(frame)==___tSUBTYPED)
    {
      /* continuation's frame is in the heap */

      if (ra == ___internal_return)
        ___RESULT = ((___WORD*)(frame-___tSUBTYPED+___WS))[1];
      else
        ___RESULT = ra;
    }
  else
    {
      /* continuation's frame is in the stack */

      if (ra == ___internal_return)
        ___RESULT = ((___WORD*)frame)[0];
      else
        ___RESULT = ra;
    }
}" f))

(define (##frame-dyn-env f)
  (##vector-ref f 2))

(define (##frame-fs f)
  (##c-code "
{
  ___WORD f = ___ARG1;
  ___WORD frame = ___VECTORREF(f,___FIX(0));
  ___WORD ra    = ___VECTORREF(f,___FIX(1));
  int fs;

  if (___TYP(frame)==___tSUBTYPED)
    {
      /* continuation's frame is in the heap */

      if (ra == ___internal_return)
        {
          ___WORD fs_link = ((___WORD *)(((___WORD*)(frame-___tSUBTYPED+___WS))[1]-___tPROCEDURE))[0];
          fs = (fs_link>>3)&((1<<14)-1);
        }
      else
        {
          ___WORD fs_link = ((___WORD *)(ra-___tPROCEDURE))[0];
          fs = (fs_link>>3)&((1<<14)-1);
        }
    }
  else
    {
      /* continuation's frame is in the stack */

      if (ra == ___internal_return)
        {
          ___WORD fs_link = ((___WORD *)(((___WORD*)frame)[0]-___tPROCEDURE))[0];
          fs = (fs_link>>3)&((1<<14)-1);
        }
      else
        {
          ___WORD fs_link = ((___WORD *)(ra-___tPROCEDURE))[0];
          fs = (fs_link>>3)&((1<<14)-1);
        }
    }

  ___RESULT = ___FIX(fs);
}" f))

(define (##frame-stk-ref f i)
  (##c-code "
{
  ___WORD f = ___ARG1;
  int i = ___INT(___ARG2);
  ___WORD frame = ___VECTORREF(f,___FIX(0));
  ___WORD ra    = ___VECTORREF(f,___FIX(1));
  int fs;

  if (___TYP(frame)==___tSUBTYPED)
    {
      /* continuation's frame is in the heap */

      if (ra == ___internal_return)
        {
          ___WORD fs_link = ((___WORD *)(((___WORD*)(frame-___tSUBTYPED+___WS))[1]-___tPROCEDURE))[0];
          fs = (fs_link>>3)&((1<<14)-1);
        }
      else
        {
          ___WORD fs_link = ((___WORD *)(ra-___tPROCEDURE))[0];
          fs = (fs_link>>3)&((1<<14)-1);
        }
      ___RESULT = ((___WORD*)(frame-___tSUBTYPED+___WS))[fs-i+1];  /* what if i==link and frame is first in section???? */
    }
  else
    {
      /* continuation's frame is in the stack */

      if (ra == ___internal_return)
        {
          ___WORD fs_link = ((___WORD *)(((___WORD*)frame)[0]-___tPROCEDURE))[0];
          fs = (fs_link>>3)&((1<<14)-1);
        }
      else
        {
          ___WORD fs_link = ((___WORD *)(ra-___tPROCEDURE))[0];
          fs = (fs_link>>3)&((1<<14)-1);
        }
      ___RESULT = ((___WORD*)frame)[fs-i];  /* what if i==link and frame is first in section???? */
    }

}" f i))

(define (##frame-next f)
  (##c-code "
{
  ___WORD f = ___ARG1;
  int i = ___INT(___ARG2);
  ___WORD frame = ___VECTORREF(f,___FIX(0));
  ___WORD ra    = ___VECTORREF(f,___FIX(1));
  ___WORD denv  = ___VECTORREF(f,___FIX(2));
  ___WORD *fp, frame_ra, next_ra, next_frame;
  int fs, link;

  if (___TYP(frame)==___tSUBTYPED)
    {
      /* continuation's frame is in the heap */

      fp = (___WORD*)(frame-___tSUBTYPED+___WS);

      if (ra == ___internal_return)
        {
          ___WORD fs_link = ((___WORD *)(fp[1]-___tPROCEDURE))[0];
          fs = (fs_link>>3)&((1<<14)-1);
          link = (fs_link>>17)&((1<<14)-1);
        }
      else
        {
          ___WORD fs_link = ((___WORD *)(ra-___tPROCEDURE))[0];
          fs = (fs_link>>3)&((1<<14)-1);
          link = (fs_link>>17)&((1<<14)-1);
        }
      fp += fs+1;

      next_ra = fp[-link-1];
      next_frame = fp[-fs-1];

      if (next_frame == 0)
        ___RESULT = ___FAL;
      else
        {
          ___hp-=((((3)+1)*___WS+___AU-1)&-___AU)/___WS;
          {___WORD *___ptr=___hp+1;
          *___hp=((3)<<___SB+___TB)+(___sCONTFRAME<<___TB);
          ___STORE(next_frame)
          ___STORE(next_ra)
          ___STORE(denv)}
          ___RESULT = ___GET_VECTOR;
        }
    }
  else
    {
      /* continuation's frame is in the stack */

      if (ra == ___internal_return)
        {
          ___WORD fs_link = ((___WORD *)(((___WORD*)frame)[0]-___tPROCEDURE))[0];
          fs = (fs_link>>3)&((1<<14)-1);
          link = (fs_link>>17)&((1<<14)-1);
        }
      else
        {
          ___WORD fs_link = ((___WORD *)(ra-___tPROCEDURE))[0];
          fs = (fs_link>>3)&((1<<14)-1);
          link = (fs_link>>17)&((1<<14)-1);
        }
      fp = ((___WORD*)frame)+fs;
      frame_ra = fp[-link-1];

      if (frame_ra == ___ps->handler_break)
        {
          /* first frame of that section */

          next_ra    = fp[0];
          next_frame = fp[1];
        }
      else
        {
          /* not the first frame of that section */

          next_ra    = frame_ra;
          next_frame = (___WORD)fp;
        }

      if (next_frame == 0)
        ___RESULT = ___FAL;
      else
        {
          ___hp-=((((3)+1)*___WS+___AU-1)&-___AU)/___WS;
          {___WORD *___ptr=___hp+1;
          *___hp=((3)<<___SB+___TB)+(___sCONTFRAME<<___TB);
          ___STORE(next_frame)
          ___STORE(next_ra)
          ___STORE(denv)}
          ___RESULT = ___GET_VECTOR;
        }
    }
}" f))

(define (##closure? x)
  (##c-code "
{
  if (___TYP(___ARG1)==___tPROCEDURE&&___TYP(*(___WORD *)((___ARG1)-___tPROCEDURE))==___tPROCEDURE)
    ___RESULT = ___TRU;
  else
    ___RESULT = ___FAL;
}" x))

(define (##subprocedure? x)
  (##c-code "
{
  if ((___TYP(___ARG1)==___tPROCEDURE) &&
      ( (*((___WORD *)(___ARG1-___tPROCEDURE))&7) > 2))
    ___RESULT = ___TRU;
  else
    ___RESULT = ___FAL;
}" x))

(define (##subprocedure-id x)
  (##c-code "
{
  if (___TYP(___ARG1)==___tPROCEDURE)
  {
    ___WORD *start = ((___WORD *)(___ARG1-___tPROCEDURE));
    ___WORD *ptr = start;
    while (*ptr != 0) ptr -= 2;
    ptr += 2;
    ___RESULT = ___FIX( (start-ptr)/2 );
  }
  else
    ___RESULT = ___FIX(0);
}" x))

(define (##subprocedure-parent x)
  (##c-code "
{
  if (___TYP(___ARG1)==___tPROCEDURE)
  {
    ___WORD *start = ((___WORD *)(___ARG1-___tPROCEDURE));
    ___WORD *ptr = start;
    while (*ptr != 0) ptr -= 2;
    ptr += 2;
   ___RESULT = (___WORD)ptr + ___tPROCEDURE;
  }
  else
    ___RESULT = ___FAL;
}" x))

(define (##procedure-info x)
  #f)

(define ##dynamic-environment '())

(define (##dynamic-env-bind dyn-env thunk)
  (let ((env ##dynamic-environment))
    (set! ##dynamic-environment dyn-env)
    (let ((val (thunk)))
      (set! ##dynamic-environment env)
      val)))

(define (##dynamic-env-ref)
  ##dynamic-environment)

;------------------------------------------------------------------------------

; Global variables

(define (##make-global-var id)
  (##c-code "
if (___VECTORREF(___ARG1,___FIX(2)) == 0)
{
___glo_struct *p = (___glo_struct*)malloc(sizeof(___glo_struct));
if (p == 0)
  ___RESULT = ___FAL;
else
{
  p->value = ___UNB;
  p->next = ___ps->glo_list;
  ___ps->glo_list = p;
  ___VECTORSET(___ARG1,___FIX(2),(___WORD)p)
  ___RESULT = ___ARG1;
}
}
else
  ___RESULT = ___ARG1;
" id))

(define (##object->global-var->identifier val)
  (let loop1 ((i (##fixnum.- (##vector-length ##symbol-table) 1)) (res #f))
    (if (##fixnum.< i 0)
      res
      (let loop2 ((l (##vector-ref ##symbol-table i)) (res res))
        (if (##null? l)
          (loop1 (##fixnum.- i 1) res)
          (loop2 (##cdr l)
                 (##c-code
"
if ((___VECTORREF(___ARG2,___FIX(2)) != 0) &&
    (*(___WORD*)___VECTORREF(___ARG2,___FIX(2)) == ___ARG1) &&
    ((___ARG3 == ___FAL) ||
     (___VECTORREF(___ARG2,___FIX(2)) < ___VECTORREF(___ARG3,___FIX(2)))))
  ___RESULT = ___ARG2;
else
  ___RESULT = ___ARG3;
" val (##car l) res)))))))

(define (##global-var->identifier gv)
  gv)

;------------------------------------------------------------------------------

; Symbols

(define (##make-uninterned-symbol name) ; name must be a non-mutable string or symbol
  (##subtype-set! (##vector name #f 0) 1))

;------------------------------------------------------------------------------

; Miscelaneous

(define (##switch-task-internal . lst)
  #f)

(define (##cpu-times vect)
  (##c-code "
___VECTORSET(___ARG1,___FIX(0L),___FIX(os_clock()))
___RESULT = ___ARG1;
" vect))

(define (##touch-undetermined ph thunk)
  (let ((val (##touch (thunk))))
    (##c-code "
if (___UNBOX(___ARG1,___tPLACEHOLDER)[0] == ___ARG1)
  ___UNBOX(___ARG1,___tPLACEHOLDER)[0] = ___ARG2;
___RESULT = ___UNBOX(___ARG1,___tPLACEHOLDER)[0];
" ph val)))

(define (##make-queue)
  (##cons '() '()))

(define (##make-semaphore . lst)
  (##err "##make-semaphore"))

(define (##os-file-block-read . lst)
  (##err "##os-file-block-read"))

(define (##os-file-block-write . lst)
  (##err "##os-file-block-write"))

(define (##os-get-next-event . lst)
  (##err "##os-get-next-event"))

(define (##os-handle-event . lst)
  (##err "##os-handle-event"))

(define (##os-set-timer-interval interval)
  #f)

(define (##queue-get! . lst)
  #f)

(define (##queue-peek-list . lst)
  #f)

(define (##queue-put! . lst)
  #f)

(define (##quit)
  (##c-code "exit(0);"))

(define (##real-time . lst)
  (##err "##real-time"))

(define (##semaphore-signal . lst)
  (##err "##semaphore-signal"))

(define (##semaphore-wait . lst)
  (##err "##semaphore-wait"))

(define (##sequentially thunk)
  (thunk))

(define ##default-ed-file "noname.scm")

(define (ed (filename ##default-ed-file))
  (if (##string? filename)
    (begin
      (set! ##default-ed-file filename)
      (##shell-command (##string-append "editor " filename))
      filename)
    (##trap-check-string 'ed filename)))

(define (##shell-command (cmd #f))
  (let ((command (if (##string? cmd) cmd "sh")))
    (##c-code "
___RESULT = ___FIX(os_shell_command((char *)(___ARG1-___tSUBTYPED+___WS)));
" command)))

(define ##object-file-count 0)

(define (##load-object-file name)
  (let ((n (##fixnum.+ ##object-file-count 1)))
    (set! ##object-file-count n)
    (let* ((filename
             (##string-append name ".O"))
           (new-filename
             (##string-append "./temp.O_" (##number->string n 10)))
           (res1
             (##shell-command
               (##string-append "cp " filename " " new-filename)))
           (x (##c-code "
___W_ALL gc( ___pstate, 0 ); ___R_ALL /* to prevent gc while loading */
___W_ALL
if (load_object_file((char *)(___ARG1-___tSUBTYPED+___WS)))
  ___RESULT = ___FAL;
else
  ___RESULT = *___module_tbl[nb_modules-1]->lp+2*___WS;
___R_ALL
" new-filename)))
      (##shell-command
        (##string-append "rm " new-filename))
      (if x
        x
        "(unknown error)"))))

(define ##cc-program
  ((##c-procedure () string "___result = CC_PROGRAM;")))

(define ##ld-program
  ((##c-procedure () string "___result = LD_PROGRAM;")))

(declare (separate))

(define c#cf #f) ; no compiler!
(define c#compile-file-dynamic #f)

(define (compile-file-dynamic filename . options)
  (touch-vars (filename)
    (check-string filename (compile-file-dynamic filename . options)
      (let ((p c#compile-file-dynamic))
        (if (##procedure? p)
          (##apply p (##cons filename options))
          (##runtime-error
            "Compiler is not available"
            'compile-file-dynamic
            (##cons filename options)))))))

(define ##gc-report #f)

(define (##gc)
  (##c-code "
___W_ALL gc( ___pstate, 0 ); ___R_ALL
")
  #f)


; execute each module in sequence

(define (##initial-continuation)
  (let loop ((i 1))
    (let ((ev ##exec-vector))
      (let ((len (##vector-length ev)))
        (if (##fixnum.< i len)
          (if (##fixnum.= i (##fixnum.- len 1))
            ((##vector-ref ev i))
            (begin
              ((##vector-ref ev i))
              (loop (##fixnum.+ i 1))))))))
  (##quit))

(##initial-continuation)
