;==============================================================================

; file: "_nonstd.scm"

; Copyright (C) 1994, 1995 by Marc Feeley, All Rights Reserved.

(##include "header.scm")

;------------------------------------------------------------------------------

; Non-standard procedures

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(##define-macro (define-runtime-macro . rest)
  `(##eval-global '(##define-macro ,@rest)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (exit)
  (##quit))

(define (error msg . args)
  (##call-with-current-continuation (lambda (cont) (##sequentially (lambda ()
    (##identify-error #f msg args '())
    (##debug-repl cont))))))

(define (eval expr (env))
  (##eval-global expr))

(define (compile-file filename . options)
  (touch-vars (filename)
    (check-string filename (compile-file filename . options)
      (let ((cf c#cf))
        (if (##procedure? cf)
          (##apply cf (##cons filename (##cons #f options)))
          (##runtime-error
            "Compiler is not available"
            'compile-file
            (##cons filename options)))))))
  
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define-runtime-macro (define-macro pattern body . rest)
  `(##DEFINE-MACRO ,pattern ,body ,@rest))

(define-runtime-macro (declare . rest)
  `(##DECLARE ,@rest))

(define-runtime-macro (include filename)
  `(##INCLUDE ,filename))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define-runtime-macro (dynamic-define var val)
  `(##DYNAMIC-DEFINE ',var ,val))

(define-runtime-macro (dynamic-ref var)
  `(##DYNAMIC-REF ',var))

(define-runtime-macro (dynamic-set! var val)
  `(##DYNAMIC-SET! ',var ,val))

(define-runtime-macro (dynamic-let bindings body . rest)
  (##dynamic-let-build (##cons bindings (##cons body rest))))

(define (##dynamic-let-build src)

  (define (build src bindings)
    (if (##pair? bindings)
      (let ((binding (##touch-list (##car bindings))))
        (##shape src binding 2)
        (let ((x (##car binding)))
          (touch-vars (x)
            (let ((rest (build src (##cdr bindings))))
              (if (##not (##symbol? x))
                (##signal '##SIGNAL.SYNTAX-ERROR src "Identifier expected:" x))
              (##cons (##list '##CONS (##list 'QUOTE x) (##cadr binding))
                      rest)))))
      (if (##null? bindings)
        '()
        (##signal '##SIGNAL.SYNTAX-ERROR src "Ill-terminated bindings"))))

  (##list '##DYNAMIC-LET
          (##cons '##LIST (build src (##car src)))
          (##cons 'LAMBDA (##cons '() (##cdr src)))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define-runtime-macro (define-structure name . field-defs)

  (define (err)
    (##signal '##SIGNAL.SYNTAX-ERROR
              (##cons 'define-structure (##cons name field-defs))
              "Ill-formed special form:"
              'define-structure))

  (define (sym . strings)
    (##string->symbol (##apply ##string-append strings)))

  (if (##symbol? name)
    (let* ((name-str (##symbol->string name))
           (tag (sym "##structure-tag." name-str)))

      (define (add-field i field fields)
        (let* ((field-str (##symbol->string field))
               (field-ref (sym name-str "-" field-str))
               (field-set! (sym name-str "-" field-str "-set!")))
          (##cons `(SET! ,field-set!
                     (LAMBDA (X Y)
                       (##STRUCTURE-FIELD-SET! X Y ,tag ,i)))
                  (##cons `(SET! ,field-ref
                             (LAMBDA (X)
                               (##STRUCTURE-FIELD-REF X ,tag ,i)))
                          fields))))

      (define (generate-definitions field-ops all-fields printed-fields)
        (let ((params (##reverse all-fields)))
          `(BEGIN
             (SET! ,tag '#(,name ,@(##reverse printed-fields)))
             ,@field-ops
             (SET! ,(sym "make-" name-str)
               (LAMBDA (,@params)
                 (##SUBTYPE-SET! (##VECTOR ,tag ,@params) 9)))
             (SET! ,(sym name-str "?")
               (LAMBDA (X)
                 (##STRUCTURE-TEST X ,tag)))
             ',name)))

      (let loop1 ((l1 field-defs) (l2 '()) (l3 '()) (i 1))
        (if (##pair? l1)
          (let ((rest (##cdr l1)) (field (##car l1)))
            (cond ((##symbol? field)
                   (loop1 rest
                         (add-field i field l2)
                         (##cons field l3)
                         (##fixnum.+ i 1)))
                  ((and (or (##null? field) (##pair? field))
                        (##not (##pair? rest)))
                   (let ((printed-fields l3))
                     (let loop2 ((l1 field) (l2 l2) (l3 l3) (i i))
                       (if (##pair? l1)
                         (let ((rest (##cdr l1)) (field (##car l1)))
                           (cond ((##symbol? field)
                                  (loop2 rest
                                        (add-field i field l2)
                                        (##cons field l3)
                                        (##fixnum.+ i 1)))
                                  (else
                                   (err))))
                         (generate-definitions l2 l3 printed-fields)))))
                  (else
                   (err))))
          (generate-definitions l2 l3 l3))))
    (err)))

(define (##structure-test obj tag)
  (and (##structure? obj)
       (##eq? (##vector-ref obj 0) tag)))

(define (##structure-field-ref obj tag i)
  (if (and (##structure? obj)
           (##eq? (##vector-ref obj 0) tag))
    (##vector-ref obj i)
    (##structure-error tag i #f(##list obj))))

(define (##structure-field-set! obj v tag i)
  (if (and (##structure? obj)
           (##eq? (##vector-ref obj 0) tag))
    (##vector-set! obj i v)
    (##structure-error tag i #t (##list obj v))))

(define (##structure-error tag i set!? args)
  (let ((name-str (##symbol->string (##vector-ref tag 0)))
        (field-str (##symbol->string (##vector-ref tag i))))
    (##runtime-error (##string-append "Structure of type `"
                                      name-str
                                      "' expected")
                     (##string->symbol
                       (if set!?
                         (##string-append name-str "-" field-str "-set!")
                         (##string-append name-str "-" field-str)))
                     args)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define-runtime-macro (trace . procs)

  (define (tr l)
    (if (##pair? l)
      (let ((var (##car l)))
        (##cons (##list '##TRACE
                        (##list 'QUOTE var)
                        (##list 'LAMBDA '() var)
                        (##list 'LAMBDA '(##VAL) (##list 'SET! var '##VAL)))
                (tr (##cdr l))))
      '()))

  (if (##pair? procs)
    (##cons 'BEGIN (tr procs))
    '(##TRACE-LIST)))

(define-runtime-macro (untrace . procs)

  (define (untr l)
    (if (##pair? l)
      (let ((var (##car l)))
        (##cons (##list '##UNTRACE (##list 'QUOTE var)) (untr (##cdr l))))
      '()))

  (if (##pair? procs)
    (##cons 'BEGIN (untr procs))
    '(##UNTRACE-ALL)))

(define ##traced '())

(define (##trace name getter setter)

  (define (add-quotes l)
    (if (##pair? l)
      (let ((x (##car l)))
        (##cons (if (##self-eval? x) x (##list 'QUOTE x))
                (add-quotes (##cdr l))))
      '()))

  (define (traced-proc proc)
    (lambda args
      (let* ((i (##dynamic-ref '##TRACE-INDENT 0))
             (w (if (##fixnum.< 40 i) 0 (##fixnum.- 40 i)))
             (out (##repl-out))
             (call (##cons name (add-quotes args))))

        (define (indent i)
          (let loop ((j 0))
            (if (##fixnum.< j i)
              (begin
                (##write-string (if (##fixnum.= (##fixnum.remainder j 3) 0) "|" " ") out)
                (loop (##fixnum.+ j 1))))))

        (indent i)
        (##write-string "Entry " out)
        (##write-string (##object->string call (##fixnum.+ w 33) (if-touches #t #f)) out)
        (##newline out)
        (let ((result
                (##dynamic-let
                  (##list (##cons '##TRACE-INDENT (##fixnum.+ i 1)))
                  (lambda () (##apply proc args)))))
          (indent i)
          (##write-string "==> " out)
          (##write-string (##object->string result (##fixnum.+ w 35) (if-touches #t #f)) out)
          (##newline out)
          result))))

  (let ((proc (getter)))
    (if (##procedure? proc)
      (let ((x (##assq name ##traced)))
        (if (##not (and x (##eq? proc (##cadddr x)))) ; being traced already?
          (let ((tproc (traced-proc proc)))
            (if x
              (begin
                (if (##eq? ((##cadr x)) (##cadddr x)) ; var = traced proc?
                  ((##caddr x) (##car (##cddddr x)))) ; restore old value
                (##set-car! (##cdr x) getter)
                (##set-car! (##cddr x) setter)
                (##set-car! (##cdddr x) tproc)
                (##set-car! (##cddddr x) proc))
              (set! ##traced
                (##cons (##list name getter setter tproc proc) ##traced)))
            (setter tproc))))))
  name)

(define (##trace-list)
  (let loop ((l1 ##traced) (l2 '()))
    (if (##pair? l1)
      (let ((x (##car l1)))
        (loop (##cdr l1) (##cons (##car x) l2)))
      l2)))

(define (##untrace name)
  (let loop ((l1 ##traced) (l2 '()))
    (if (##pair? l1)
      (let ((x (##car l1)))
        (if (##eq? (##car x) name)
          (begin
            (if (##eq? ((##cadr x)) (##cadddr x)) ; var = traced proc?
              ((##caddr x) (##car (##cddddr x)))) ; restore old value
            (set! ##traced (##append (##reverse l2) (##cdr l1)))
            name)
          (loop (##cdr l1) (##cons x l2))))
      name)))

(define (##untrace-all)
  (let loop ((l ##traced))
    (if (##pair? l)
      (let ((x (##car l)))
        (##untrace (##car x))
        (loop (##cdr l)))
      ##unprint-object)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (set-gc-report report?)
  (set! ##gc-report report?)
  ##unprint-object)

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (open-input-string s)
  (touch-vars (s)
    (check-string s (open-input-string s)
      (##open-input-string s))))

(define (open-output-string)
  (##open-output-string))

(define (get-output-string p)
  (touch-vars (p)
    (check-output-port p (get-output-string p)
      (check-open-port p (get-output-string p)
        (##get-output-string p)))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (with-input-from-string s thunk)
  (touch-vars (s thunk)
    (check-string s (with-input-from-string s thunk)
      (check-procedure thunk (with-input-from-string s thunk)
        (let ((port (##open-input-string s)))
          (##dynamic-let
            (##list (##cons '##CURRENT-INPUT-PORT port))
            thunk))))))

(define (with-output-to-string thunk)
  (touch-vars (thunk)
    (check-procedure thunk (with-output-to-string thunk)
      (let ((port (##open-output-string)))
        (##dynamic-let
          (##list (##cons '##CURRENT-OUTPUT-PORT port))
          thunk)
        (##get-output-string port)))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (with-input-from-port port thunk)
  (touch-vars (port thunk)
    (check-input-port port (with-input-from-port port thunk)
      (check-open-port port (with-input-from-port port thunk)
        (check-procedure thunk (with-input-from-port port thunk)
          (##dynamic-let (##list (##cons '##CURRENT-INPUT-PORT port)) thunk))))))

(define (with-output-to-port port thunk)
  (touch-vars (port thunk)
    (check-output-port port (with-output-to-port port thunk)
      (check-open-port port (with-output-to-port port thunk)
        (check-procedure thunk (with-output-to-port port thunk)
          (##dynamic-let (##list (##cons '##CURRENT-OUTPUT-PORT port)) thunk))))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (pretty-print obj (p) (w))

  (define (pretty-print* obj port width)
    (##pretty-print obj port width)
    ##unprint-object)

  (if (##unassigned? p)
    (let ((port (##current-output-port)))
      (check-open-port port (pretty-print obj)
        (pretty-print* obj port (##port-width port))))
    (touch-vars (p)
      (if (##unassigned? w)
        (check-output-port p (pretty-print obj p)
          (check-open-port p (pretty-print obj p)
            (pretty-print* obj p (##port-width p))))
        (touch-vars (w)
          (check-output-port p (pretty-print obj p w)
            (check-open-port p (pretty-print obj p w)
              (check-exact-int-non-neg w (pretty-print obj p w)
                (pretty-print* obj p w)))))))))

(define (pp obj (p) (w))

  (define (pp* obj port width)
    (if (##procedure? obj)
      (##pretty-print (##decompile obj) port width)
      (##pretty-print obj port width))
    ##unprint-object)

  (if (##unassigned? p)
    (let ((port (##current-output-port)))
      (check-open-port port (pp obj)
        (pp* obj port (##port-width port))))
    (touch-vars (p)
      (if (##unassigned? w)
        (check-output-port p (pp obj p)
          (check-open-port p (pp obj p)
            (pp* obj p (##port-width p))))
        (touch-vars (w)
          (check-output-port p (pp obj p w)
            (check-open-port p (pp obj p w)
              (check-exact-int-non-neg w (pp obj p w)
                (pp* obj p w)))))))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (runtime)
  (let ((buf (##vector 0 0)))
    (##cpu-times buf)
    (##exact->inexact (##/ (##vector-ref buf 0) 1000))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (gensym (prefix 'g))
  (touch-vars (prefix)
    (check-symbol prefix (gensym prefix)
      (##make-uninterned-symbol prefix))))

;------------------------------------------------------------------------------
