/* file: "gambit.h" */

/* Copyright (C) 1994, 1995 by Marc Feeley, All Rights Reserved. */

/*---------------------------------------------------------------------------*/

/* DETERMINE MACHINE CHARACTERISTICS */

/*
 * The following symbols need to be defined to reflect the
 * machine's characteristics and those of the C compiler:
 *
 *   define ___MACH_32_BIT_LONG if the machine has 32 bit "long" integers
 *   define ___MACH_64_BIT_LONG if the machine has 64 bit "long" integers
 *
 *   define ___MACH_BIG_ENDIAN    if the machine is "big endian"
 *   define ___MACH_LITTLE_ENDIAN if the machine is "little endian"
 *
 */

#ifdef sparc
#define ___MACH_32_BIT_LONG
#define ___MACH_BIG_ENDIAN
#endif

#ifdef sgi
#define ___MACH_32_BIT_LONG
#define ___MACH_BIG_ENDIAN
#endif

#ifdef __alpha
#define ___MACH_64_BIT_LONG
#define ___MACH_LITTLE_ENDIAN
#endif

#ifdef hpux
#define ___MACH_32_BIT_LONG
#define ___MACH_BIG_ENDIAN
#endif

#ifdef _AIX
#define ___MACH_32_BIT_LONG
#define ___MACH_BIG_ENDIAN
#endif

#ifdef i386
#define ___MACH_32_BIT_LONG
#define ___MACH_LITTLE_ENDIAN
#endif

#ifdef _X86_
#define ___MACH_32_BIT_LONG
#define ___MACH_LITTLE_ENDIAN
#endif

#ifdef m68k
#define ___MACH_32_BIT_LONG
#define ___MACH_BIG_ENDIAN
#endif

#ifdef __MWERKS__
#define ___MACH_32_BIT_LONG
#define ___MACH_BIG_ENDIAN
#endif

#ifdef MACHTEN
#define ___MACH_32_BIT_LONG
#define ___MACH_BIG_ENDIAN
#endif

/*---------------------------------------------------------------------------*/

/* PREFERENCES */

/*
 * These definitions can be altered to improve performance.
 */

/* Is it better to grow stack up (towards positive) or down (towards 0)? */

#ifndef ___STK_GROW_UP
#ifndef ___STK_GROW_DOWN
#define ___STK_GROW_DOWN
#endif
#endif

/* Is it better to adjust stack incrementally or in jumps? */

#ifndef ___STK_GROW_JUMP
#ifndef ___STK_GROW_INCR
#define ___STK_GROW_INCR
#endif
#endif

/* If a "switch" has less than ___SMALL_SWITCH_LIMIT cases, it will  */
/* be turned into a cascade of "if"s.  This transformation is turned */
/* off by setting ___SMALL_SWITCH_LIMIT to 0.                        */

#ifndef ___SMALL_SWITCH_LIMIT
#define ___SMALL_SWITCH_LIMIT 8
#endif

/* Generate a single or multiple C host procedure(s) per Scheme module? */

#ifndef ___SINGLE_HOST
#ifndef ___MULTIPLE_HOSTS
#define ___MULTIPLE_HOSTS
#endif
#endif

/* Compile for static linking or dynamic linking. */

#ifndef ___STATIC_LINK
#ifndef ___DYNAMIC_LINK
#define ___STATIC_LINK
#endif
#endif

/* >>>>>>>>>>>>> DO NOT CHANGE ANY DEFINITION AFTER THIS POINT <<<<<<<<<<<<< */

/*---------------------------------------------------------------------------*/

/* SANITY CHECK */

#ifdef ___MACH_32_BIT_LONG
#ifdef ___MACH_64_BIT_LONG
#undef ___MACH_32_BIT_LONG
#undef ___MACH_64_BIT_LONG
@error You must set either "___MACH_32_BIT_LONG" or "___MACH_64_BIT_LONG"
#endif
#else
#ifndef ___MACH_64_BIT_LONG
@error You must set either "___MACH_32_BIT_LONG" or "___MACH_64_BIT_LONG"
#endif
#endif

#ifdef ___MACH_BIG_ENDIAN
#ifdef ___MACH_LITTLE_ENDIAN
#undef ___MACH_BIG_ENDIAN
#undef ___MACH_LITTLE_ENDIAN
@error You must set either "___MACH_BIG_ENDIAN" or "___MACH_LITTLE_ENDIAN"
#endif
#else
#ifndef ___MACH_LITTLE_ENDIAN
@error You must set either "___MACH_BIG_ENDIAN" or "___MACH_LITTLE_ENDIAN"
#endif
#endif

#ifndef ___VERSION
@error "___VERSION" has not been set by includer
#endif

#if ___VERSION != 4
@error Incompatible version of "gambit.h"
#endif

/*---------------------------------------------------------------------------*/

/*
 * IMPORTANT NOTE:
 *
 * As a general principle, the macros are written in such a way that all
 * macros which expand into C expressions yield PRIMARY expressions.  Also,
 * macros assume that arguments are PRIMARY expressions.  A PRIMARY
 * expression is either:
 *
 *   an identifier, or
 *   a non-negative constant, or
 *   a parenthesized expression.
 *
 * This convention avoids many operator priority problems and helps keep
 * the number of parentheses down.  On the other hand, it is easy to
 * introduce errors when modifying these macros so be careful.
 */

/*---------------------------------------------------------------------------*/

/* GENERAL DEFINITIONS */

/*
 * ___WORD  must be an integral type able to contain a pointer with no loss
 * ___FLOAT is the numerical type for 64 bit IEEE floating-point numbers
 * ___WS    is equal to "sizeof( ___WORD )"
 * ___FW    is equal to "sizeof( ___FLOAT ) / sizeof( ___WORD )"
 * ___U8    must be an unsigned integral type of exactly 8 bits
 * ___U16   must be an unsigned integral type of exactly 16 bits
 * ___U32   must be an unsigned integral type of exactly 32 bits
 * ___S32   must be a signed integral type of exactly 32 bits
 */

#define ___WORD long
#define ___FLOAT double

#ifdef ___MACH_32_BIT_LONG
#define ___WS 4
#define ___FW 2
#else
#define ___WS 8
#define ___FW 1
#endif

#define ___U8  unsigned char
#define ___U16 unsigned short
#define ___U32 unsigned int
#define ___S32 int

/* Padding at end of objects so that they can be aligned by moving them */
/* forward.                                                             */

#ifdef ___MACH_32_BIT_LONG
#define ___PADDING ,0
#define ___PADDING_LBL ,{0,0}
#else
#define ___PADDING
#define ___PADDING_LBL
#endif

/* Utility macro to select code for MULTIPLE/SINGLE host. */

#ifdef ___MULTIPLE_HOSTS
#define ___SM(s,m)m
#else
#define ___SM(s,m)s
#endif

/*---------------------------------------------------------------------------*/

/* Object representation */

/* Type tags (___TB = number of bits used for type tags) */

/* NOTE: ___tFIXNUM must be 0 */

#define ___TB 3
#define ___tFIXNUM      0
#define ___tFLONUM      1
#define ___tSPECIAL     7
#define ___tPAIR        4
#define ___tSUBTYPED    3
#define ___tPROCEDURE   2
#define ___tPLACEHOLDER 5

/* Subtype tags (___SB = number of bits used for subtype tags) */

#define ___SB (8-___TB)
#define ___sVECTOR    0
#define ___sSYMBOL    1
#define ___sPORT      2
#define ___sRATNUM    3
#define ___sCPXNUM    4
#define ___sFRAME     5
#define ___sTASK      6
#define ___sQUEUE     7
#define ___sSEMAPHORE 8
#define ___sSTRUCTURE 9
#define ___sPOINTER   10
#define ___sCONTFRAME 11
#define ___sWEAKCELL  15
#define ___sSTRING    16
#define ___sBIGNUM    17

/* Object reference tags (must be different from ___tFIXNUM and ___tSPECIAL) */

#define ___rSYM ___tPROCEDURE
#define ___rCNS ___tPAIR
#define ___rFLO ___tFLONUM
#define ___rSUB ___tSUBTYPED

/* Value constructors for object references inside executable code */

#define ___FAL (((___WORD)-1<<___TB)+___tSPECIAL)
#define ___TRU (((___WORD)-2<<___TB)+___tSPECIAL)
#define ___UND (((___WORD)-3<<___TB)+___tSPECIAL)
#define ___UNA (((___WORD)-4<<___TB)+___tSPECIAL)
#define ___UNB (((___WORD)-5<<___TB)+___tSPECIAL)
#define ___EOF (((___WORD)-6<<___TB)+___tSPECIAL)
#define ___NUL (((___WORD)-7<<___TB)+___tSPECIAL)
#define ___BRK (((___WORD)-8<<___TB)+___tSPECIAL)
#define ___FIX(x)((___WORD)((___S32)x<<___TB))
#define ___CHR(x)(((___WORD)(___U8)x<<___TB)+___tSPECIAL)

#define ___SYM(i)(___WORD)___sym_tbl[i]
#define ___CNS(i)___BOX((___ALIGNUP(___cns_tbl)+i*2),___tPAIR)
#define ___FLO(i)___BOX((___ALIGNUP(___flo_tbl)+i*___FW),___tFLONUM)
#define ___SUB(i)(___WORD)___sub_tbl[i]

/* Value constructors for object references inside memory allocated objects */

#define ___REF_FAL ___FAL
#define ___REF_TRU ___TRU
#define ___REF_UND ___UND
#define ___REF_UNA ___UNA
#define ___REF_UNB ___UNB
#define ___REF_EOF ___EOF
#define ___REF_NUL ___NUL
#define ___REF_FIX(x)___FIX(x)
#define ___REF_CHR(x)___CHR(x)

#define ___REF_SYM(i)(((___WORD)i<<___TB)+___rSYM)
#define ___REF_CNS(i)(((___WORD)i<<___TB)+___rCNS)
#define ___REF_FLO(i)(((___WORD)i<<___TB)+___rFLO)
#define ___REF_SUB(i)(((___WORD)i<<___TB)+___rSUB)

/*---------------------------------------------------------------------------*/

/* Miscellaneous macros */

#define ___NOTHING

#define ___INT(x)(x>>___TB)

/*---------------------------------------------------------------------------*/

/* Align pointer to next multiple of allocation unit (___AU) */

#define ___AU (1<<___TB)
#define ___ALIGNUP(ptr)((___WORD*)((___WORD)ptr+___AU-1&-___AU))

/* Masks to extract tag and data part of a value */

#define ___TMASK (((___WORD)1<<___TB)-1)
#define ___DMASK (-((___WORD)1<<___TB))
#define ___TYP(x)(x&___TMASK)
#define ___DAT(x)(x&___DMASK)
#define ___TESTTYPE(x,typ)(___TYP(x)==typ)

#define ___BOX(ptr,tag)((___WORD)ptr+tag)
#define ___UNBOX(ptr,tag)((___WORD*)(ptr-tag))

/* Masks to extract subtype tag and length of a subtyped value */

#define ___SMASK ((1L<<___SB+___TB)-1)
#define ___LMASK (-(1L<<___SB+___TB))
#define ___TESTSUBTYPE(x,s)(___TYP((___temp=x))==___tSUBTYPED&&___SUBTYPE(___temp)==s<<___TB)
#define ___HEADER(x)*___UNBOX(x,___tSUBTYPED)/***parens needed?*/

/* Flonum boxing and unboxing */

#define ___FLOBOX(x)(*(___FLOAT*)(___hp-=___FW)=x,___BOX(___hp,___tFLONUM))
#define ___FLOUNBOX(x)(*(___FLOAT*)___UNBOX(x,___tFLONUM))

/*****************************************************************************/

/* Inlinable operations (for APPLY and IFJUMP gvm instructions) */

/* APPLY-able operations */

#define ___TYPE(x)___FIX(___TYP(x))
#define ___TYPECAST(x,y)((x&___DMASK)+___INT(y))
#define ___SUBTYPE(x)(___HEADER(x)&___SMASK)
#define ___SUBTYPESET(x,y)___temp=x;___HEADER(___temp)=(___HEADER(___temp)&___LMASK)+(y);

/* IFJUMP-able operations */

#define ___FALSEP(x)(x==___FAL)
#define ___NULLP(x)(x==___NUL)
#define ___UNASSIGNEDP(x)(x==___UNA)
#define ___UNBOUNDP(x)(x==___UNB)
#define ___EQP(x,y)(x==y)

/* IFJUMP-able operations */

#define ___FIXNUMP(x)___TESTTYPE(x,___tFIXNUM)
#define ___FLONUMP(x)___TESTTYPE(x,___tFLONUM)
#define ___SPECIALP(x)___TESTTYPE(x,___tSPECIAL)
#define ___PAIRP(x)___TESTTYPE(x,___tPAIR)
#define ___SUBTYPEDP(x)___TESTTYPE(x,___tSUBTYPED)
#define ___PROCEDUREP(x)___TESTTYPE(x,___tPROCEDURE)
#define ___PLACEHOLDERP(x)___TESTTYPE(x,___tPLACEHOLDER)
#define ___VECTORP(x)___TESTSUBTYPE(x,___sVECTOR)
#define ___SYMBOLP(x)___TESTSUBTYPE(x,___sSYMBOL)
#define ___RATNUMP(x)___TESTSUBTYPE(x,___sRATNUM)
#define ___CPXNUMP(x)___TESTSUBTYPE(x,___sCPXNUM)
#define ___STRINGP(x)___TESTSUBTYPE(x,___sSTRING)
#define ___BIGNUMP(x)___TESTSUBTYPE(x,___sBIGNUM)
#define ___CHARP(x)(___TYP((___temp=x))==___tSPECIAL&&___temp>=0)

/* APPLY-able operations */

#define ___FIX0 ___FIX(0)
#define ___FIXPOS(x)((___WORD)((___S32)x))
#define ___FIXADD(x,y)((___WORD)((___S32)x+(___S32)y))
#define ___FIX1 ___FIX(1)
#define ___FIXMUL(x,y)((___WORD)((___S32)x*(___S32)___INT(y)))
#define ___FIXNEG(x)((___WORD)(-(___S32)x))
#define ___FIXSUB(x,y)((___WORD)((___S32)x-(___S32)y))
#define ___FIXINV(x)___FIX(((___S32)___FIX(1)/(___S32)x))
#define ___FIXQUO(x,y)___FIX(((___S32)x/(___S32)y))
#define ___FIXREM(x,y)((___WORD)((___S32)x%(___S32)y))/****is this ok?*******/
#define ___FIXMOD(x,y)((___temp=___FIXREM((___ps->temp1=x),(___ps->temp2=y)))==0?___FIX(0):((___ps->temp1<0)==(___ps->temp2<0)?(___S32)___temp:___FIXADD(___temp,___ps->temp2)))
#define ___FIXIOR(x,y)((___WORD)((___S32)x|(___S32)y))
#define ___FIXXOR(x,y)((___WORD)((___S32)x^(___S32)y))
#define ___FIXM1 ___FIX(-1)
#define ___FIXAND(x,y)((___WORD)((___S32)x&(___S32)y))
#define ___FIXNOT(x)((___WORD)((___S32)x^(___S32)___DMASK))
#define ___FIXASH(x,y)((___temp=___INT(y))<0?((___S32)x>>-___temp)&___DMASK:(___WORD)((___S32)x<<___temp))
#define ___FIXLSH(x,y)((___temp=___INT(y))<0?((___U32)x>>-___temp)&___DMASK:(___WORD)((___S32)x<<___temp))

/* IFJUMP-able operations */

#define ___FIXZEROP(x)(x==0)
#define ___FIXPOSITIVEP(x)(x>0)
#define ___FIXNEGATIVEP(x)(x<0)
#define ___FIXODDP(x)(x&___FIX(1))
#define ___FIXEVENP(x)!(x&___FIX(1))
#define ___FIXEQ(x,y)(x==y)
#define ___FIXLT(x,y)(x<y)
#define ___FIXGT(x,y)(x>y)
#define ___FIXLE(x,y)(x<=y)
#define ___FIXGE(x,y)(x>=y)

/* APPLY-able operations */

#define ___FLOTOFIX(x)___FIX(___FLOUNBOX(x))
#define ___FLOFROMFIX(x)___FLOBOX(___INT(x))
#define ___FLO0 ___FLOBOX(0.0)
#define ___FLOPOS(x)x
#define ___FLOADD(x,y)(x+y)
#define ___FLO1 ___FLOBOX(1.0)
#define ___FLOMUL(x,y)(x*y)
#define ___FLONEG(x)___FLOBOX(-___FLOUNBOX(x))
#define ___FLOSUB(x,y)(x-y)
#define ___FLOINV(x)___FLOBOX(1.0/___FLOUNBOX(x))
#define ___FLODIV(x,y)(x/y)
#define ___FLOABS(x)___FLOBOX(fabs(___FLOUNBOX(x)))
#define ___FLOFLOOR(x)___FLOBOX(floor(___FLOUNBOX(x)))
#define ___FLOCEILING(x)___FLOBOX(ceil(___FLOUNBOX(x)))
#define ___FLOROUND(x)___FLOBOX(floor(0.5+___FLOUNBOX(x)))
#define ___FLOEXP(x)___FLOBOX(exp(___FLOUNBOX(x)))
#define ___FLOLOG(x)___FLOBOX(log(___FLOUNBOX(x)))
#define ___FLOSIN(x)___FLOBOX(sin(___FLOUNBOX(x)))
#define ___FLOCOS(x)___FLOBOX(cos(___FLOUNBOX(x)))
#define ___FLOTAN(x)___FLOBOX(tan(___FLOUNBOX(x)))
#define ___FLOASIN(x)___FLOBOX(asin(___FLOUNBOX(x)))
#define ___FLOACOS(x)___FLOBOX(acos(___FLOUNBOX(x)))
#define ___FLOATAN(x)___FLOBOX(atan(___FLOUNBOX(x)))
#define ___FLOSQRT(x)___FLOBOX(sqrt(___FLOUNBOX(x)))

/* IFJUMP-able operations */

#define ___FLOZEROP(x)(___FLOUNBOX(x)==0.0)
#define ___FLOPOSITIVEP(x)(___FLOUNBOX(x)>0.0)
#define ___FLONEGATIVEP(x)(___FLOUNBOX(x)<0.0)
#define ___FLOEQ(x,y)(___FLOUNBOX(x)==___FLOUNBOX(y))
#define ___FLOLT(x,y)(___FLOUNBOX(x)<___FLOUNBOX(y))
#define ___FLOGT(x,y)(___FLOUNBOX(x)>___FLOUNBOX(y))
#define ___FLOLE(x,y)(___FLOUNBOX(x)<=___FLOUNBOX(y))
#define ___FLOGE(x,y)(___FLOUNBOX(x)>=___FLOUNBOX(y))

/* IFJUMP-able operations */

#define ___CHAREQP(x,y)(x==y)
#define ___CHARLTP(x,y)(x<y)
#define ___CHARGTP(x,y)(x>y)
#define ___CHARLEP(x,y)(x<=y)
#define ___CHARGEP(x,y)(x>=y)

/* APPLY-able operations */

#define ___CONS(x,y)(*--___hp=x,*--___hp=y,___BOX(___hp,___tPAIR))
#define ___SETCAR(x,y)___UNBOX(x,___tPAIR)[1]=y;
#define ___SETCDR(x,y)___UNBOX(x,___tPAIR)[0]=y;
#define ___CAR(x)___UNBOX(x,___tPAIR)[1]
#define ___CDR(x)___UNBOX(x,___tPAIR)[0]
#define ___CAAR(x)___CAR(___CAR(x))
#define ___CADR(x)___CAR(___CDR(x))
#define ___CDAR(x)___CDR(___CAR(x))
#define ___CDDR(x)___CDR(___CDR(x))
#define ___CAAAR(x)___CAR(___CAR(___CAR(x)))
#define ___CAADR(x)___CAR(___CAR(___CDR(x)))
#define ___CADAR(x)___CAR(___CDR(___CAR(x)))
#define ___CADDR(x)___CAR(___CDR(___CDR(x)))
#define ___CDAAR(x)___CDR(___CAR(___CAR(x)))
#define ___CDADR(x)___CDR(___CAR(___CDR(x)))
#define ___CDDAR(x)___CDR(___CDR(___CAR(x)))
#define ___CDDDR(x)___CDR(___CDR(___CDR(x)))
#define ___CAAAAR(x)___CAR(___CAR(___CAR(___CAR(x))))
#define ___CAAADR(x)___CAR(___CAR(___CAR(___CDR(x))))
#define ___CAADAR(x)___CAR(___CAR(___CDR(___CAR(x))))
#define ___CAADDR(x)___CAR(___CAR(___CDR(___CDR(x))))
#define ___CADAAR(x)___CAR(___CDR(___CAR(___CAR(x))))
#define ___CADADR(x)___CAR(___CDR(___CAR(___CDR(x))))
#define ___CADDAR(x)___CAR(___CDR(___CDR(___CAR(x))))
#define ___CADDDR(x)___CAR(___CDR(___CDR(___CDR(x))))
#define ___CDAAAR(x)___CDR(___CAR(___CAR(___CAR(x))))
#define ___CDAADR(x)___CDR(___CAR(___CAR(___CDR(x))))
#define ___CDADAR(x)___CDR(___CAR(___CDR(___CAR(x))))
#define ___CDADDR(x)___CDR(___CAR(___CDR(___CDR(x))))
#define ___CDDAAR(x)___CDR(___CDR(___CAR(___CAR(x))))
#define ___CDDADR(x)___CDR(___CDR(___CAR(___CDR(x))))
#define ___CDDDAR(x)___CDR(___CDR(___CDR(___CAR(x))))
#define ___CDDDDR(x)___CDR(___CDR(___CDR(___CDR(x))))

/* APPLY-able operations */

#define ___MAKECELL(x)___CONS(x,___NUL)
#define ___CELLREF(x)___UNBOX(x,___tPAIR)[1]
#define ___CELLSET(x,y)___UNBOX(x,___tPAIR)[1]=y;

#define ___VECTORLENGTH(x)___DAT(___HEADER(x)>>___SB)
#define ___VECTORREF(x,y)(___UNBOX(x,___tSUBTYPED)+1)[___INT(y)]
#define ___VECTORSET(x,y,z)(___UNBOX(x,___tSUBTYPED)+1)[___INT(y)]=z;
#define ___VECTORSHRINK(x,y)___temp=x;___HEADER(___temp)=(___HEADER(___temp)&(1<<___SB+___TB)-1)+((y)<<___SB);

#define ___STRINGLENGTH(x)(___DAT(___HEADER(x)>>___SB)-___FIX(1))
#define ___STRINGREF(x,y)___CHR(*(___U8*)((x)-___tSUBTYPED+___WS+(___INT(y))))
#define ___STRINGSET(x,y,z)*(___U8*)((x)-___tSUBTYPED+___WS+(___INT(y)))=___INT(z);
#define ___STRINGSHRINK(x,y)___temp=x;___ps->temp2=y;___HEADER(___temp)=(___HEADER(___temp)&(1<<___SB+___TB)-1)+(___ps->temp2+___FIX(1)<<___SB);*(___U8*)(___temp-___tSUBTYPED+___WS+(___INT(___ps->temp2)))=0;

#define ___VECTOR8LENGTH(x)___DAT(___HEADER(x)>>___SB)
#define ___VECTOR8REF(x,y)___FIX(*(___U8*)((x)-___tSUBTYPED+___WS+(___INT(y))))
#define ___VECTOR8SET(x,y,z)*(___U8*)((x)-___tSUBTYPED+___WS+(___INT(y)))=___INT(z);
#define ___VECTOR8SHRINK(x,y)___temp=x;___HEADER(___temp)=(___HEADER(___temp)&(1<<___SB+___TB)-1)+((y)<<___SB);

#define ___VECTOR16LENGTH(x)___DAT(___HEADER(x)>>___SB+1)
#define ___VECTOR16REF(x,y)___FIX(*(___U16*)((x)-___tSUBTYPED+___WS+((y)>>___TB-1)))
#define ___VECTOR16SET(x,y,z)*(___U16*)((x)-___tSUBTYPED+___WS+((y)>>___TB-1))=___INT(z);
#define ___VECTOR16SHRINK(x,y)___temp=x;___HEADER(___temp)=(___HEADER(___temp)&(1<<___SB+___TB)-1)+((y)<<___SB+1);

#define ___CLOSURECODE(x)___CLO(x,-1)
#define ___CLOSUREREF(x,y)___CLO(x,___INT(y))
#define ___CLOSURESET(x,y,z)___SET_CLO(x,___INT(y),z)

#define ___GLOBALVARREF(gv)((___glo_struct*)___VECTORREF(gv,___FIX(2)))->value
#define ___GLOBALVARSET(gv,val)((___glo_struct*)___VECTORREF(gv,___FIX(2)))->value = val;

#define ___MAKEPLACEHOLDER(x)(*--___hp=x,--___hp,*___hp=___BOX(___hp,___tPLACEHOLDER))

#define ___PSTATE (___WORD)___ps

/*---------------------------------------------------------------------------*/

/* Stack manipulation */

#ifdef ___STK_GROW_DOWN

#define ___PUSH(val)*--___sp=(val);

#ifdef ___STK_GROW_INCR
#define ___PSSTK(fpo,spo)___PSSP[-(spo)]
#define ___PSADJFPSP(fpa,spa)___PSSP-=(spa);
#define ___POP(fpo)*___sp++
#define ___SET_PUSH(fpo,val)*--___sp=(val);
#define ___STK(fpo,spo)___sp[-(spo)]
#define ___SET_STK(fpo,spo,val)___sp[-(spo)]=(val);
#define ___ADJFP(fpa)
#define ___ADJSP(spa)___sp-=(spa);
#define ___ADJFPSP(fpa,spa)___sp-=(spa);
#else
#define ___PSSTK(fpo,spo)___PSSP[-(fpo)]
#define ___PSADJFPSP(fpa,spa)___PSSP-=(fpa);
#define ___POP(fpo)___sp[-(fpo)]
#define ___SET_PUSH(fpo,val)___sp[-(fpo)]=(val);
#define ___STK(fpo,spo)___sp[-(fpo)]
#define ___SET_STK(fpo,spo,val)___sp[-(fpo)]=(val);
#define ___ADJFP(fpa)___sp-=(fpa);
#define ___ADJSP(spa)
#define ___ADJFPSP(fpa,spa)___sp-=(fpa);
#endif

#else

#define ___PUSH(val)*___sp++=(val);

#ifdef ___STK_GROW_INCR
#define ___PSSTK(fpo,spo)___PSSP[(spo)-1]
#define ___PSADJFPSP(fpa,spa)___PSSP+=(spa);
#define ___POP(fpo)*--___sp
#define ___SET_PUSH(fpo,spo)*___sp++=(spo);
#define ___STK(fpo,spo)___sp[(spo)-1]
#define ___SET_STK(fpo,spo,val)___sp[(spo)-1]=(val);
#define ___ADJFP(fpa)
#define ___ADJSP(spa)___sp+=(spa);
#define ___ADJFPSP(fpa,spa)___sp+=(spa);
#else
#define ___PSSTK(fpo,spo)___PSSP[(fpo)-1]
#define ___PSADJFPSP(fpa,spa)___PSSP+=(fpa);
#define ___POP(fpo)___sp[(fpo)-1]
#define ___SET_PUSH(fpo,spo)___sp[(fpo)-1]=(spo);
#define ___STK(fpo,spo)___sp[(fpo)-1]
#define ___SET_STK(fpo,spo,val)___sp[(fpo)-1]=(val);
#define ___ADJFP(fpa)___sp+=(fpa);
#define ___ADJSP(spa)
#define ___ADJFPSP(fpa,spa)___sp+=(fpa);
#endif

#endif

#define ___HOST_PROC ___SM(___MH_PROC,___PH_PROC)
#define ___HOST_LBL0 ___SM(1,___PH_LBL0)

#define ___BEGIN_M_COD ___SM(___BEGIN_COD,___NOTHING)
#define ___END_M_COD ___SM(___END_COD,___NOTHING)
#define ___BEGIN_P_COD ___SM(___NOTHING,___BEGIN_COD)
#define ___END_P_COD ___SM(___NOTHING,___END_COD)

#define ___BEGIN_P_SLBL
#define ___DEF_P_SLBL(id)
#define ___END_P_SLBL

#define ___D_ALL ___SM(___MD_ALL,___PD_ALL)
#define ___R_ALL ___SM(___MR_ALL,___PR_ALL)
#define ___W_ALL ___SM(___MW_ALL,___PW_ALL)

#define ___BEGIN_COD \
static ___WORD ___HOST_PROC (___pstate)___processor ___pstate;{ \
register ___processor ___ps=___pstate; \
register ___WORD ___pc, ___temp; \
register ___WORD ___start=___lp+(___HOST_LBL0<<___LB); \
___D_ALL ___R_ALL

#define ___END_COD \
___W_ALL return ___pc;}

#ifdef ___MACH_32_BIT_LONG
#define ___WB 2
#else
#define ___WB 3
#endif
#define ___LB (1+___WB)

#define ___PSHEAP ___ps->hp
#define ___D_HEAP register ___WORD *___hp;
#define ___R_HEAP ___hp=___PSHEAP;
#define ___W_HEAP ___PSHEAP=___hp;

#define ___PSSP ___ps->sp
#define ___D_SP register ___WORD *___sp;
#define ___R_SP ___sp=___PSSP;
#define ___W_SP ___PSSP=___sp;

/*---------------------------------------------------------------------------*/

/* GVM registers */

#define ___MAX_NB_GVM_REGS 8
#define ___NB_GVM_REGS 5
#define ___NB_ARG_REGS 3

#define ___PSR0 ___ps->r[0]
#define ___D_R0 register ___WORD ___r0;
#define ___R_R0 ___r0=___PSR0;
#define ___W_R0 ___PSR0=___r0;
#define ___R0 ___r0
#define ___SET_R0(val)___r0=val;

#define ___PSR1 ___ps->r[1]
#define ___D_R1 register ___WORD ___r1;
#define ___R_R1 ___r1=___PSR1;
#define ___W_R1 ___PSR1=___r1;
#define ___R1 ___r1
#define ___SET_R1(val)___r1=(val);

#define ___PSR2 ___ps->r[2]
#define ___D_R2 register ___WORD ___r2;
#define ___R_R2 ___r2=___PSR2;
#define ___W_R2 ___PSR2=___r2;
#define ___R2 ___r2
#define ___SET_R2(val)___r2=(val);

#define ___PSR3 ___ps->r[3]
#define ___D_R3 register ___WORD ___r3;
#define ___R_R3 ___r3=___PSR3;
#define ___W_R3 ___PSR3=___r3;
#define ___R3 ___r3
#define ___SET_R3(val)___r3=(val);

#define ___PSR4 ___ps->r[4]
#define ___D_R4 register ___WORD ___r4;
#define ___R_R4 ___r4=___PSR4;
#define ___W_R4 ___PSR4=___r4;
#define ___R4 ___r4
#define ___SET_R4(val)___r4=(val);

#define ___PSR5 ___ps->r[5]
#define ___D_R5 register ___WORD ___r5;
#define ___R_R5 ___r5=___PSR5;
#define ___W_R5 ___PSR5=___r5;
#define ___R5 ___r5
#define ___SET_R5(val)___r5=(val);

#define ___PSR6 ___ps->r[6]
#define ___D_R6 register ___WORD ___r6;
#define ___R_R6 ___r6=___PSR6;
#define ___W_R6 ___PSR6=___r6;
#define ___R6 ___r6
#define ___SET_R6(val)___r6=(val);

#define ___PSR7 ___ps->r[7]
#define ___D_R7 register ___WORD ___r7;
#define ___R_R7 ___r7=___PSR7;
#define ___W_R7 ___PSR7=___r7;
#define ___R7 ___r7
#define ___SET_R7(val)___r7=(val);

/*---------------------------------------------------------------------------*/

/* Labels and switches */

#define ___NOT(x)!x
#define ___AND(x,y)(x&&y)

#define ___LBL(n)___start+((___PH_LBL0-___HOST_LBL0+n)<<___LB)
#define ___DEF_SLBL(n,lbl)case ___PH_LBL0-___HOST_LBL0+n:___DEF_GLBL(lbl)
#define ___DEF_GLBL(lbl)lbl:
#define ___GOTO(lbl)goto lbl;
#define ___IF_GOTO(test,lbl)if(test)___GOTO(lbl)
#define ___JUMPLOCAL(nargs,val,lbl)___SM(___GOTO(lbl),{nargs ___pc=(val);goto ___jump;})
#define ___JUMP(nargs,val){nargs ___pc=(val);goto ___jump;}
#define ___JUMPEXT(nargs,val){nargs ___pc=(val);goto ___jumpext;}

/* Big switch */

#ifdef ___MULTIPLE_HOSTS

#define ___BEGIN_M_SW
#define ___END_M_SW
#define ___BEGIN_M_SW_CLO
#define ___END_M_SW_CLO
#define ___BEGIN_P_SW ___BEGIN_SW
#define ___END_P_SW ___END_SW
#define ___BEGIN_P_SW_CLO ___BEGIN_SW_CLO
#define ___END_P_SW_CLO ___END_SW_CLO

#else

#define ___BEGIN_M_SW ___BEGIN_SW
#define ___END_M_SW ___END_SW
#define ___BEGIN_M_SW_CLO ___BEGIN_SW_CLO
#define ___END_M_SW_CLO ___END_SW_CLO
#define ___BEGIN_P_SW
#define ___END_P_SW
#define ___BEGIN_P_SW_CLO
#define ___END_P_SW_CLO

#undef ___SMALL_SWITCH_LIMIT
#define ___SMALL_SWITCH_LIMIT 0

#endif

#define ___BEGIN_SW \
___pc=___ps->pc;___jump:___pc-=___start;if((___pc&___TMASK)==0)switch(___pc>>___LB){default:break;
#define ___END_SW \
}___pc+=___start;___jumpext:

#define ___BEGIN_SW_CLO \
___pc=___ps->pc;___jump:___pc-=___start;if((___pc&___TMASK)==0)switch(___pc>>___LB){default:{___WORD ___orig=___pc+___start;if (((___WORD*)(___orig-___tPROCEDURE))[1]==(___WORD)___HOST_PROC){___SET_R4(___orig);___pc=((___WORD*)(___orig-___tPROCEDURE))[0];goto ___jump;}break;}
#define ___END_SW_CLO \
}___pc+=___start;___jumpext:

/* Small switch */

#if ___SMALL_SWITCH_LIMIT < 1
#define ___DEF_SLBL1(n,lbl)___DEF_SLBL(n,lbl)
#define ___BEGIN_P_SW1(a,aa)___BEGIN_P_SW
#define ___END_P_SW1 ___END_P_SW
#define ___BEGIN_P_SW_CLO1(a,aa)___BEGIN_P_SW_CLO
#define ___END_P_SW_CLO1 ___END_P_SW_CLO
#else
#define ___DEF_SLBL1(n,lbl)
#define ___BEGIN_P_SW1(a,aa)
#define ___END_P_SW1 ___jump:___jumpext:
#define ___BEGIN_P_SW_CLO1(a,aa)
#define ___END_P_SW_CLO1 ___jump:___jumpext:
#endif

#if ___SMALL_SWITCH_LIMIT < 2
#define ___DEF_SLBL2(n,lbl)___DEF_SLBL(n,lbl)
#define ___BEGIN_P_SW2(a,aa,b,bb)___BEGIN_P_SW
#define ___END_P_SW2 ___END_P_SW
#define ___BEGIN_P_SW_CLO2(a,aa,b,bb)___BEGIN_P_SW_CLO
#define ___END_P_SW_CLO2 ___END_P_SW_CLO
#else
#define ___DEF_SLBL2(n,lbl)lbl:
#define ___BEGIN_P_SW2(a,aa,b,bb)___SS_HEAD \
___SS(a,aa)___SS(b,bb)___SS_TAIL
#define ___END_P_SW2 ___jumpext:
#define ___BEGIN_P_SW_CLO2(a,aa,b,bb)___SS_HEAD \
___SS(a,aa)___SS(b,bb)___SS_TAIL_CLO
#define ___END_P_SW_CLO2 ___jumpext:
#endif

#if ___SMALL_SWITCH_LIMIT < 3
#define ___DEF_SLBL3(n,lbl)___DEF_SLBL(n,lbl)
#define ___BEGIN_P_SW3(a,aa,b,bb,c,cc)___BEGIN_P_SW
#define ___END_P_SW3 ___END_P_SW
#define ___BEGIN_P_SW_CLO3(a,aa,b,bb,c,cc)___BEGIN_P_SW_CLO
#define ___END_P_SW_CLO3 ___END_P_SW_CLO
#else
#define ___DEF_SLBL3(n,lbl)lbl:
#define ___BEGIN_P_SW3(a,aa,b,bb,c,cc)___SS_HEAD \
___SS(a,aa)___SS(b,bb)___SS(c,cc)___SS_TAIL
#define ___END_P_SW3 ___jumpext:
#define ___BEGIN_P_SW_CLO3(a,aa,b,bb,c,cc)___SS_HEAD \
___SS(a,aa)___SS(b,bb)___SS(c,cc)___SS_TAIL_CLO
#define ___END_P_SW_CLO3 ___jumpext:
#endif

#if ___SMALL_SWITCH_LIMIT < 4
#define ___DEF_SLBL4(n,lbl)___DEF_SLBL(n,lbl)
#define ___BEGIN_P_SW4(a,aa,b,bb,c,cc,d,dd)___BEGIN_P_SW
#define ___END_P_SW4 ___END_P_SW
#define ___BEGIN_P_SW_CLO4(a,aa,b,bb,c,cc,d,dd)___BEGIN_P_SW_CLO
#define ___END_P_SW_CLO4 ___END_P_SW_CLO
#else
#define ___DEF_SLBL4(n,lbl)lbl:
#define ___BEGIN_P_SW4(a,aa,b,bb,c,cc,d,dd)___SS_HEAD \
___SS(a,aa)___SS(b,bb)___SS(c,cc)___SS(d,dd)___SS_TAIL
#define ___END_P_SW4 ___jumpext:
#define ___BEGIN_P_SW_CLO4(a,aa,b,bb,c,cc,d,dd)___SS_HEAD \
___SS(a,aa)___SS(b,bb)___SS(c,cc)___SS(d,dd)___SS_TAIL_CLO
#define ___END_P_SW_CLO4 ___jumpext:
#endif

#if ___SMALL_SWITCH_LIMIT < 5
#define ___DEF_SLBL5(n,lbl)___DEF_SLBL(n,lbl)
#define ___BEGIN_P_SW5(a,aa,b,bb,c,cc,d,dd,e,ee)___BEGIN_P_SW
#define ___END_P_SW5 ___END_P_SW
#define ___BEGIN_P_SW_CLO5(a,aa,b,bb,c,cc,d,dd,e,ee)___BEGIN_P_SW_CLO
#define ___END_P_SW_CLO5 ___END_P_SW_CLO
#else
#define ___DEF_SLBL5(n,lbl)lbl:
#define ___BEGIN_P_SW5(a,aa,b,bb,c,cc,d,dd,e,ee)___SS_HEAD \
___SS(a,aa)___SS(b,bb)___SS(c,cc)___SS(d,dd)___SS(e,ee)___SS_TAIL
#define ___END_P_SW5 ___jumpext:
#define ___BEGIN_P_SW_CLO5(a,aa,b,bb,c,cc,d,dd,e,ee)___SS_HEAD \
___SS(a,aa)___SS(b,bb)___SS(c,cc)___SS(d,dd)___SS(e,ee)___SS_TAIL_CLO
#define ___END_P_SW_CLO5 ___jumpext:
#endif

#if ___SMALL_SWITCH_LIMIT < 6
#define ___DEF_SLBL6(n,lbl)___DEF_SLBL(n,lbl)
#define ___BEGIN_P_SW6(a,aa,b,bb,c,cc,d,dd,e,ee,f,ff)___BEGIN_P_SW
#define ___END_P_SW6 ___END_P_SW
#define ___BEGIN_P_SW_CLO6(a,aa,b,bb,c,cc,d,dd,e,ee,f,ff)___BEGIN_P_SW_CLO
#define ___END_P_SW_CLO6 ___END_P_SW_CLO
#else
#define ___DEF_SLBL6(n,lbl)lbl:
#define ___BEGIN_P_SW6(a,aa,b,bb,c,cc,d,dd,e,ee,f,ff)___SS_HEAD \
___SS(a,aa)___SS(b,bb)___SS(c,cc)___SS(d,dd)___SS(e,ee)___SS(f,ff) \
___SS_TAIL
#define ___END_P_SW6 ___jumpext:
#define ___BEGIN_P_SW_CLO6(a,aa,b,bb,c,cc,d,dd,e,ee,f,ff)___SS_HEAD \
___SS(a,aa)___SS(b,bb)___SS(c,cc)___SS(d,dd)___SS(e,ee)___SS(f,ff) \
___SS_TAIL_CLO
#define ___END_P_SW_CLO6 ___jumpext:
#endif

#if ___SMALL_SWITCH_LIMIT < 7
#define ___DEF_SLBL7(n,lbl)___DEF_SLBL(n,lbl)
#define ___BEGIN_P_SW7(a,aa,b,bb,c,cc,d,dd,e,ee,f,ff,g,gg)___BEGIN_P_SW
#define ___END_P_SW7 ___END_P_SW
#define ___BEGIN_P_SW_CLO7(a,aa,b,bb,c,cc,d,dd,e,ee,f,ff,g,gg)___BEGIN_P_SW_CLO
#define ___END_P_SW_CLO7 ___END_P_SW_CLO
#else
#define ___DEF_SLBL7(n,lbl)lbl:
#define ___BEGIN_P_SW7(a,aa,b,bb,c,cc,d,dd,e,ee,f,ff,g,gg)___SS_HEAD \
___SS(a,aa)___SS(b,bb)___SS(c,cc)___SS(d,dd)___SS(e,ee)___SS(f,ff)___SS(g,gg) \
___SS_TAIL
#define ___END_P_SW7 ___jumpext:
#define ___BEGIN_P_SW_CLO7(a,aa,b,bb,c,cc,d,dd,e,ee,f,ff,g,gg)___SS_HEAD \
___SS(a,aa)___SS(b,bb)___SS(c,cc)___SS(d,dd)___SS(e,ee)___SS(f,ff)___SS(g,gg) \
___SS_TAIL_CLO
#define ___END_P_SW_CLO7 ___jumpext:
#endif

#if ___SMALL_SWITCH_LIMIT < 8
#define ___DEF_SLBL8(n,lbl)___DEF_SLBL(n,lbl)
#define ___BEGIN_P_SW8(a,aa,b,bb,c,cc,d,dd,e,ee,f,ff,g,gg,h,hh)___BEGIN_P_SW
#define ___END_P_SW8 ___END_P_SW
#define ___BEGIN_P_SW_CLO8(a,aa,b,bb,c,cc,d,dd,e,ee,f,ff,g,gg,h,hh)___BEGIN_P_SW_CLO
#define ___END_P_SW_CLO8 ___END_P_SW_CLO
#else
#define ___DEF_SLBL8(n,lbl)lbl:
#define ___BEGIN_P_SW8(a,aa,b,bb,c,cc,d,dd,e,ee,f,ff,g,gg,h,hh)___SS_HEAD \
___SS(a,aa)___SS(b,bb)___SS(c,cc)___SS(d,dd)___SS(e,ee)___SS(f,ff)___SS(g,gg) \
___SS(h,hh)___SS_TAIL
#define ___END_P_SW8 ___jumpext:
#define ___BEGIN_P_SW_CLO8(a,aa,b,bb,c,cc,d,dd,e,ee,f,ff,g,gg,h,hh)___SS_HEAD \
___SS(a,aa)___SS(b,bb)___SS(c,cc)___SS(d,dd)___SS(e,ee)___SS(f,ff)___SS(g,gg) \
___SS(h,hh)___SS_TAIL_CLO
#define ___END_P_SW_CLO8 ___jumpext:
#endif

#define ___SS(n,lbl)if(___pc==((n)<<___LB))goto lbl;
#define ___SS_HEAD ___pc=___ps->pc;___jump:___pc-=___start;
#define ___SS_TAIL ___pc+=___start;goto ___jumpext;
#define ___SS_TAIL_CLO {___WORD ___orig=___pc+___start;if (((___WORD*)(___orig-___tPROCEDURE))[1]==(___WORD)___HOST_PROC){___SET_R4(___orig);___pc=((___WORD*)(___orig-___tPROCEDURE))[0];goto ___jump;}___pc=___orig;goto ___jumpext;}

/*---------------------------------------------------------------------------*/

#ifdef ___DYNAMIC_LINK

#define ___GLO(i,glo)___glo_tbl[i]->value
#define ___SET_GLO(i,glo,val)___glo_tbl[i]->value=val;

#else

#define ___GLO(i,glo)glo.value
#define ___SET_GLO(i,glo,val)glo.value=val;

#endif

#define ___PRC(i)___start+(i-___HOST_LBL0<<___LB)

#define ___SET_NARGS(n)___ps->na=n;
#define ___IF_NARGS_EQ(n,code)if(___ps->na==n){code}else
#define ___WRONG_NARGS(lbl){___ps->temp1=___LBL(lbl);___JUMPEXT(___NOTHING,___ps->handler_wrong_nargs)}
#define ___WRONG_NARGS_CLO ___JUMPEXT(___NOTHING,___ps->handler_wrong_nargs_clo)
#define ___BUILD_REST(lbl)if(___ps->na>=0){___ps->temp1=___LBL(lbl);___JUMPEXT(___NOTHING,___ps->handler_build_rest)}
#define ___BUILD_REST_CLO if(___ps->na>=0)___JUMPEXT(___NOTHING,___ps->handler_build_rest_clo)

#define ___BOOL(x)(x)?___TRU:___FAL

#define ___ALLOC_CLO(n)(___hp-=((((n)+2)*___WS+___AU-1)&-___AU)/___WS,((___WORD)___hp)+___tPROCEDURE)
#define ___SETUP_CLO(clo,lbl,stores){___WORD *___ptr=(___WORD*)((clo)-___tPROCEDURE); *___ptr++=___LBL(lbl); *___ptr++=(___WORD)___HOST_PROC; stores}
#define ___CLO(x,y)*(___WORD*)((x)-___tPROCEDURE+___WS*(1+(y)))
#define ___SET_CLO(x,y,z)*(___WORD*)((x)-___tPROCEDURE+___WS*(1+(y)))=(z);

#define ___ALLOC_VECTOR(n,stores) \
___hp-=((((n)+1)*___WS+___AU-1)&-___AU)/___WS; \
{___WORD *___ptr=___hp+1; \
*___hp=((n)<<___SB+___TB)+(___sVECTOR<<___TB); \
stores}
#define ___GET_VECTOR (((___WORD)___hp)+___tSUBTYPED)

#define ___ALLOC_STRING(n,stores) \
___hp-=((((n)+1)+___WS+___AU-1)&-___AU)/___WS; \
{___U8 *___ptr=(___U8*)(___hp+1); \
*___hp=(((n)+1)<<___SB+___TB)+(___sSTRING<<___TB); \
stores}
#define ___GET_STRING (((___WORD)___hp)+___tSUBTYPED)

#define ___ALLOC_VECTOR8(n,stores) \
___hp-=(((n)+___WS+___AU-1)&-___AU)/___WS; \
{___U8 *___ptr=(___U8*)(___hp+1); \
*___hp=((n)<<___SB+___TB)+(___sSTRING<<___TB); \
stores}
#define ___GET_VECTOR8 (((___WORD)___hp)+___tSUBTYPED)

#define ___ALLOC_VECTOR16(n,stores) \
___hp-=((2*(n)+___WS+___AU-1)&-___AU)/___WS; \
{___U16 *___ptr=(___U16*)(___hp+1); \
*___hp=(2*(n)<<___SB+___TB)+(___sSTRING<<___TB); \
stores}
#define ___GET_VECTOR16 (((___WORD)___hp)+___tSUBTYPED)

#define ___STORE(val)*___ptr++=(val);
#define ___STOREINT(val)*___ptr++=((val)>>___TB);
#define ___STORECHAR(val)*___ptr++=((val)>>___TB);

#ifdef ___STK_GROW_DOWN
#define ___POLL(n)if(___sp<___ps->sl){___ps->temp1=___LBL(n);___JUMPEXT(___NOTHING,___ps->handler_stack_limit)}
#else
#define ___POLL(n)if(___sp>___ps->sl){___ps->temp1=___LBL(n);___JUMPEXT(___NOTHING,___ps->handler_stack_limit)}
#endif

#define ___TASK_PUSH(i)
#define ___TASK_POP(n)

#define ___CHECK_HEAP(n,m)if(___hp<___ps->hl){___ps->temp1=___LBL(n);___JUMPEXT(___NOTHING,___ps->handler_heap_limit)}

#define ___TOUCH1(n,src) \
if (___TYP((___temp=(src)))==___tPLACEHOLDER){___ps->temp1=___LBL(n);___ps->temp2=___temp;___JUMPEXT(___NOTHING,___ps->handler_touch);
#define ___TOUCH2 ___temp=___ps->temp2;}
#define ___TOUCH3 ___temp

#define ___TYPEDEFS \
typedef struct ___glo_struct { ___WORD value; struct ___glo_struct *next; } ___glo_struct; \
typedef struct \
{ \
___WORD *sl, *sp, *sb, *stack, *stack_base; \
___WORD *hl, *hp, *heap; \
___WORD r[___MAX_NB_GVM_REGS], pc, temp1, temp2, temp3; \
int na; \
int intr_enabled, intr_flag[2]; \
___WORD \
handler_break, handler_stack_limit, handler_heap_limit, \
handler_wrong_nargs, handler_wrong_nargs_clo, \
handler_build_rest,handler_build_rest_clo, \
handler_touch, handler_c_conv_error; \
___glo_struct *glo_list; \
} ___processor_struct; \
typedef ___processor_struct *___processor; \
typedef struct { ___WORD flags; ___WORD (*host)(); } ___lbl_struct; \
typedef struct \
{ \
char *name; \
___glo_struct **glo_tbl; int glo_count; char **glo_names; \
___WORD **sym_tbl; int sym_count; char **sym_names; \
___WORD *lp; \
___lbl_struct *lbl_tbl; int lbl_count; \
___WORD *cns_tbl; int cns_count; \
___WORD *flo_tbl; int flo_count; \
___WORD **sub_tbl; int sub_count; \
void (*init_proc)(); \
} ___module_struct; \
typedef ___module_struct *___module;

#define ___BEGIN_MODULE \
___TYPEDEFS \
extern double fabs(); \
extern double floor(); \
extern double ceil(); \
extern double exp(); \
extern double log(); \
extern double sin(); \
extern double cos(); \
extern double tan(); \
extern double asin(); \
extern double acos(); \
extern double atan(); \
extern double sqrt(); \
extern int ___scmobj_to_char(); \
extern int ___scmobj_to_schar(); \
extern int ___scmobj_to_uchar(); \
extern int ___scmobj_to_short(); \
extern int ___scmobj_to_ushort(); \
extern int ___scmobj_to_int(); \
extern int ___scmobj_to_uint(); \
extern int ___scmobj_to_long(); \
extern int ___scmobj_to_ulong(); \
extern int ___scmobj_to_float(); \
extern int ___scmobj_to_double(); \
extern int ___scmobj_to_pointer(); \
extern int ___scmobj_to_boolean(); \
extern int ___scmobj_to_string(); \
extern int ___char_to_scmobj(); \
extern int ___schar_to_scmobj(); \
extern int ___uchar_to_scmobj(); \
extern int ___short_to_scmobj(); \
extern int ___ushort_to_scmobj(); \
extern int ___int_to_scmobj(); \
extern int ___uint_to_scmobj(); \
extern int ___long_to_scmobj(); \
extern int ___ulong_to_scmobj(); \
extern int ___float_to_scmobj(); \
extern int ___double_to_scmobj(); \
extern int ___pointer_to_scmobj(); \
extern int ___boolean_to_scmobj(); \
extern int ___string_to_scmobj(); \
static ___WORD ___lp; \
___GLO_DECL \
___SYM_DECL

#ifdef ___DYNAMIC_LINK

#define ___DEMAND_GLO(id)
#define ___SUPPLY_GLO(id)
#define ___DEMAND_SYM(id)

#define ___BEGIN_GLO static char *___glo_names[]={
#define ___DEF_GLO(str)str,
#define ___END_GLO 0};

#define ___BEGIN_SYM static char *___sym_names[]={
#define ___DEF_SYM(id,str)str,
#define ___END_SYM 0};

#else

#define ___DEMAND_GLO(id)extern ___glo_struct id;
#define ___SUPPLY_GLO(id)extern ___glo_struct id;
#define ___DEMAND_SYM(id)extern ___WORD id[];

#define ___BEGIN_GLO
#define ___DEF_GLO(str)
#define ___END_GLO

#define ___BEGIN_SYM static ___WORD *___sym_tbl[]={
#define ___DEF_SYM(id,str)id,
#define ___END_SYM 0};

#endif

#define ___BEGIN_CNS \
static ___WORD ___cns_tbl[]={
#define ___DEF_CNS(car,cdr)cdr,car
#define ___END_CNS ___PADDING};

#define ___BEGIN_FLO \
static ___WORD ___flo_tbl[]={
#define ___DEF_FLO(val)val
#define ___END_FLO ___PADDING};

#define ___DEF_SUB_VEC(id,n) \
static ___WORD id[]={(n<<___SB+___TB)+(___sVECTOR<<___TB)

#define ___DEF_SUB_STR(id,n) \
static ___WORD id[]={(n+1<<___SB+___TB)+(___sSTRING<<___TB)

#define ___DEF_SUB_BIG(id,n) \
static ___WORD id[]={((2*n)<<___SB+___TB)+(___sBIGNUM<<___TB)

#define ___DEF_SUB_RAT(id,num,den) \
static ___WORD id[]={ \
(2<<___SB+___TB)+(___sRATNUM<<___TB),num,den ___PADDING};

#define ___DEF_SUB_CPX(id,real,imag) \
static ___WORD id[]={ \
(2<<___SB+___TB)+(___sCPXNUM<<___TB),real,imag ___PADDING};

#define ___BEGIN_SUB \
static ___WORD *___sub_tbl[]={
#define ___DEF_SUB(id)id
#define ___END_SUB };

#define ___VEC0 ___PADDING};
#define ___VEC1(a),a

#if ___WS == 4

#ifdef ___MACH_BIG_ENDIAN
#define ___S4(a,b,c,d)(((((___WORD)a)<<8)+((___WORD)b)<<8)+((___WORD)c)<<8)+((___WORD)d)
#else
#define ___S4(d,c,b,a)(((((___WORD)a)<<8)+((___WORD)b)<<8)+((___WORD)c)<<8)+((___WORD)d)
#endif

#define ___STR0 ,___S4(0,0,0,0) ___PADDING};
#define ___STR1(a),___S4(a,0,0,0) ___PADDING};
#define ___STR2(a,b),___S4(a,b,0,0) ___PADDING};
#define ___STR3(a,b,c),___S4(a,b,c,0) ___PADDING};
#define ___STR4(a,b,c,d),___S4(a,b,c,d),___S4(0,0,0,0) ___PADDING};
#define ___STR5(a,b,c,d,e),___S4(a,b,c,d),___S4(e,0,0,0) ___PADDING};
#define ___STR6(a,b,c,d,e,f),___S4(a,b,c,d),___S4(e,f,0,0) ___PADDING};
#define ___STR7(a,b,c,d,e,f,g),___S4(a,b,c,d),___S4(e,f,g,0) ___PADDING};
#define ___STR8(a,b,c,d,e,f,g,h),___S4(a,b,c,d),___S4(e,f,g,h)

#else

#ifdef ___MACH_BIG_ENDIAN
#define ___S8(a,b,c,d,e,f,g,h)((((((((((___WORD)a)<<8)+((___WORD)b)<<8)+((___WORD)c)<<8)+((___WORD)d)<<8)+((___WORD)e)<<8)+((___WORD)f)<<8)+((___WORD)g)<<8)+((___WORD)h))
#else
#define ___S8(h,g,f,e,d,c,b,a)((((((((((___WORD)a)<<8)+((___WORD)b)<<8)+((___WORD)c)<<8)+((___WORD)d)<<8)+((___WORD)e)<<8)+((___WORD)f)<<8)+((___WORD)g)<<8)+((___WORD)h))
#endif

#define ___STR0 ,___S8(0,0,0,0,0,0,0,0) ___PADDING};
#define ___STR1(a),___S8(a,0,0,0,0,0,0,0) ___PADDING};
#define ___STR2(a,b),___S8(a,b,0,0,0,0,0,0) ___PADDING};
#define ___STR3(a,b,c),___S8(a,b,c,0,0,0,0,0) ___PADDING};
#define ___STR4(a,b,c,d),___S8(a,b,c,d,0,0,0,0) ___PADDING};
#define ___STR5(a,b,c,d,e),___S8(a,b,c,d,e,0,0,0) ___PADDING};
#define ___STR6(a,b,c,d,e,f),___S8(a,b,c,d,e,f,0,0) ___PADDING};
#define ___STR7(a,b,c,d,e,f,g),___S8(a,b,c,d,e,f,g,0) ___PADDING};
#define ___STR8(a,b,c,d,e,f,g,h),___S8(a,b,c,d,e,f,g,h)

#endif

#if ___FW == 1

#define ___FLO2(a,b)((___WORD)a<<32)+(___U32)b

#else

#ifdef ___MACH_BIG_ENDIAN
#define ___FLO2(a,b)a,b
#else
#define ___FLO2(b,a)a,b
#endif

#endif

#if ___WS == 4

#ifdef ___MACH_BIG_ENDIAN
#define ___B2(a,b)(((___WORD)a)<<16)+((___WORD)b)
#else
#define ___B2(b,a)(((___WORD)a)<<16)+((___WORD)b)
#endif

#define ___BIG0 ___PADDING};
#define ___BIG1(a),___B2(a,0) ___PADDING};
#define ___BIG2(a,b),___B2(a,b) ___PADDING};
#define ___BIG3(a,b,c),___B2(a,b),___B2(c,0) ___PADDING};
#define ___BIG4(a,b,c,d),___B2(a,b),___B2(c,d)

#else

#ifdef ___MACH_BIG_ENDIAN
#define ___B4(a,b,c,d)(((((___WORD)a)<<16)+((___WORD)b)<<16)+((___WORD)c)<<16)+((___WORD)d)
#else
#define ___B4(d,c,b,a)(((((___WORD)a)<<16)+((___WORD)b)<<16)+((___WORD)c)<<16)+((___WORD)d)
#endif

#define ___BIG0 ___PADDING};
#define ___BIG1(a),___B4(a,0,0,0) ___PADDING};
#define ___BIG2(a,b),___B4(a,b,0,0) ___PADDING};
#define ___BIG3(a,b,c),___B4(a,b,c,0) ___PADDING};
#define ___BIG4(a,b,c,d),___B4(a,b,c,d)

#endif

#define ___BEGIN_LBL \
static ___lbl_struct ___lbl_tbl[]={
#define ___DEF_LBL_INFO(id){0,0}
#define ___DEF_LBL_PROC(id,nbp,nbc){(nbc<<17)+(nbp<<3)+1,___SM(___MH_PROC,id)}
#define ___DEF_LBL_RETN(id,fs,link){(link<<17)+(fs<<3)+3,___SM(___MH_PROC,id)}
#define ___DEF_LBL_RETT(id,fs,link){(link<<17)+(fs<<3)+4,___SM(___MH_PROC,id)}
#define ___DEF_LBL_RETI(id,fs,link){(link<<17)+((fs+6)<<3)+7,___SM(___MH_PROC,id)}
#define ___END_LBL ___PADDING_LBL};

#ifdef ___GLO_COUNT
#ifdef ___DYNAMIC_LINK
#define ___GLO_DECL static ___glo_struct *___glo_tbl[___GLO_COUNT];
#define ___GLO_TBL ___glo_tbl,___GLO_COUNT,___glo_names
#else
#define ___GLO_DECL
#define ___GLO_TBL 0,0,0
#endif
#else
#define ___GLO_DECL
#define ___GLO_TBL 0,0,0
#endif

#ifdef ___SYM_COUNT
#ifdef ___DYNAMIC_LINK
#define ___SYM_DECL static ___WORD *___sym_tbl[___SYM_COUNT];
#define ___SYM_TBL ___sym_tbl,___SYM_COUNT,___sym_names
#else
#define ___SYM_DECL
#define ___SYM_TBL ___sym_tbl,___SYM_COUNT,0
#endif
#else
#define ___SYM_DECL
#define ___SYM_TBL 0,0,0
#endif

#ifdef ___LBL_COUNT
#define ___LBL_TBL &___lp,___lbl_tbl,___LBL_COUNT
#else
#define ___LBL_TBL 0,0,0
#endif

#ifdef ___CNS_COUNT
#define ___CNS_TBL ___cns_tbl,___CNS_COUNT
#else
#define ___CNS_TBL 0,0
#endif

#ifdef ___FLO_COUNT
#define ___FLO_TBL ___flo_tbl,___FLO_COUNT
#else
#define ___FLO_TBL 0,0
#endif

#ifdef ___SUB_COUNT
#define ___SUB_TBL ___sub_tbl,___SUB_COUNT
#else
#define ___SUB_TBL 0,0
#endif

#ifdef ___DYNAMIC_LINK
#undef ___MODULE_LINKER
#define ___MODULE_LINKER ___init_module
#endif

#define ___BEGIN_INI static void ___init_proc(){
#define ___DEF_PRM(i,glo,n)___GLO(i,glo)=___lp+(n<<___LB);
#define ___END_INI } static ___module_struct ___module_descr={___MODULE_NAME, \
___GLO_TBL,___SYM_TBL,___LBL_TBL,___CNS_TBL,___FLO_TBL,___SUB_TBL, \
___init_proc}; \
___module ___MODULE_LINKER(){return &___module_descr;}

#define ___BEGIN_SYM_GLO
#define ___DEF_SYM_GLO(prevsym,sym,str,hash,glo) \
___glo_struct glo = {___FIX(hash), 0}; \
___WORD *sym[]={(___WORD*)prevsym,str,0,(___WORD*)&glo ___PADDING};
#define ___END_SYM_GLO(prevsym)___WORD **___sym_list = prevsym;

#define ___BEGIN_MOD1
#define ___DEF_MOD1(id)extern void id();
#define ___END_MOD1

#define ___BEGIN_MOD2 void (*___modules[])()={
#define ___DEF_MOD2(id)id
#define ___END_MOD2 ,0};

#define ___BEGIN_C_COD ___W_ALL {
#define ___END_C_COD } ___R_ALL

#define ___C_CONV_ERROR {___ps->temp1=___LBL(0);___JUMPEXT(___NOTHING,___ps->handler_c_conv_error)}

#define ___SCMOBJ_TO_CHAR(x,y)___scmobj_to_char(x,&y)
#define ___SCMOBJ_TO_SCHAR(x,y)___scmobj_to_schar(x,&y)
#define ___SCMOBJ_TO_UCHAR(x,y)___scmobj_to_uchar(x,&y)
#define ___SCMOBJ_TO_SHORT(x,y)___scmobj_to_short(x,&y)
#define ___SCMOBJ_TO_USHORT(x,y)___scmobj_to_ushort(x,&y)
#define ___SCMOBJ_TO_INT(x,y)___scmobj_to_int(x,&y)
#define ___SCMOBJ_TO_UINT(x,y)___scmobj_to_uint(x,&y)
#define ___SCMOBJ_TO_LONG(x,y)___scmobj_to_long(x,&y)
#define ___SCMOBJ_TO_ULONG(x,y)___scmobj_to_ulong(x,&y)
#define ___SCMOBJ_TO_FLOAT(x,y)___scmobj_to_float(x,&y)
#define ___SCMOBJ_TO_DOUBLE(x,y)___scmobj_to_double(x,&y)
#define ___SCMOBJ_TO_POINTER(x,y)___scmobj_to_pointer(x,&y)
#define ___SCMOBJ_TO_BOOLEAN(x,y)___scmobj_to_boolean(x,&y)
#define ___SCMOBJ_TO_STRING(x,y)___scmobj_to_string(x,&y)

#define ___CONVERT_TO_SCMOBJ(f,x,y) \
{___WORD ___val; if (f(x,&___val)) ___C_CONV_ERROR else y=___val;}

#define ___CHAR_TO_SCMOBJ(x,y)___CONVERT_TO_SCMOBJ(___char_to_scmobj,x,y)
#define ___SCHAR_TO_SCMOBJ(x,y)___CONVERT_TO_SCMOBJ(___schar_to_scmobj,x,y)
#define ___UCHAR_TO_SCMOBJ(x,y)___CONVERT_TO_SCMOBJ(___uchar_to_scmobj,x,y)
#define ___SHORT_TO_SCMOBJ(x,y)___CONVERT_TO_SCMOBJ(___short_to_scmobj,x,y)
#define ___USHORT_TO_SCMOBJ(x,y)___CONVERT_TO_SCMOBJ(___ushort_to_scmobj,x,y)
#define ___INT_TO_SCMOBJ(x,y)___CONVERT_TO_SCMOBJ(___int_to_scmobj,x,y)
#define ___UINT_TO_SCMOBJ(x,y)___CONVERT_TO_SCMOBJ(___uint_to_scmobj,x,y)
#define ___LONG_TO_SCMOBJ(x,y)___CONVERT_TO_SCMOBJ(___long_to_scmobj,x,y)
#define ___ULONG_TO_SCMOBJ(x,y)___CONVERT_TO_SCMOBJ(___ulong_to_scmobj,x,y)
#define ___FLOAT_TO_SCMOBJ(x,y)___CONVERT_TO_SCMOBJ(___float_to_scmobj,x,y)
#define ___DOUBLE_TO_SCMOBJ(x,y)___CONVERT_TO_SCMOBJ(___double_to_scmobj,x,y)
#define ___POINTER_TO_SCMOBJ(x,y)___CONVERT_TO_SCMOBJ(___pointer_to_scmobj,x,y)
#define ___BOOLEAN_TO_SCMOBJ(x,y)___CONVERT_TO_SCMOBJ(___boolean_to_scmobj,x,y)
#define ___STRING_TO_SCMOBJ(x,y)___CONVERT_TO_SCMOBJ(___string_to_scmobj,x,y)
