/* file: "gc.c" */

/* Copyright (C) 1994, 1995 by Marc Feeley, All Rights Reserved. */

/* This is the Gambit-C garbage collector */

#define ___VERSION 4
#include "gambit.h"
___BEGIN_MODULE

___DEMAND_GLO(___G__23_23gc_2dreport)

#include <stdio.h>

#include "setup.h"

___processor ps;

/*---------------------------------------------------------------------------*/

/* Define SANITY_CHECKS if you want the GC to check consistency of the heap. */

#define SANITY_CHECKS

/*
  The procedure "fatal_error" is called to indicate that an unrecoverable
  error has occured.  An explicatory message is printed on standard output
  and the program is terminated.
*/

void fatal_error (msg)
char *msg;
{
  printf (msg);
  exit (1);
}

/* Allocation pointers */

___WORD *old_space_bot, *old_space_top;
___WORD *new_space_bot, *new_space_top;
___WORD *bot, *top;
___WORD flo_ptr, flo_lim;

#ifdef SANITY_CHECKS

int scanned_region;
___WORD scanned_object;

#define CONSTANT    1
#define REGISTER    2
#define GLOBAL_VAR  3
#define STACK       4
#define PAIR        5
#define PLACEHOLDER 6
#define CLOSURE     7
#define SUBTYPED    8

void check_non_new (val)
___WORD val;
{
  int i;
  if ((val>=(___WORD)new_space_bot) && (val<(___WORD)new_space_top))
  {
    printf ("*** ERROR -- Inconsistency detected (pointer to new space)\n");
    printf ("*** Value = 0x%016lx found ",val);
    switch (scanned_region)
    {
      case CONSTANT:
        printf ("in constant area\n");
        break;
      case REGISTER:
        printf ("in register\n");
        break;
      case GLOBAL_VAR:
        printf ("in global variable\n");
        break;
      case STACK:
        printf ("in stack\n");
        break;
      case PAIR:
        printf ("in pair object\n");
        break;
      case PLACEHOLDER:
        printf ("in placeholder object\n");
        break;
      case CLOSURE:
        printf ("in closure object 0x%016lx\n",scanned_object);
        for (i=0; i<20; i++)
          printf ("*** object[%d] = 0x%016lx\n",i,((___WORD*)(scanned_object&-(1<<___TB)))[i]);
        break;
      case SUBTYPED:
        printf ("in vector-like object 0x%016lx\n",scanned_object);
        for (i=0; i<20; i++)
          printf ("*** object[%d] = 0x%016lx\n",i,((___WORD*)(scanned_object&-(1<<___TB)))[i]);
        break;
      default:
        printf ("in ???\n");
        break;
    }
    exit (1);
  }
}

void check_non_new_during_copy (val,region,object)
___WORD val;
int region;
___WORD object;
{
  int i;
  switch (___TYP (val))         /* Dispatch on type of value */
    {
    case ___tFLONUM:            /* 64 bit floating point number; boxed */
    case ___tPAIR:              /* Cons cell */
    case ___tPLACEHOLDER:       /* Placeholder */
    case ___tSUBTYPED:          /* Object with subtype (vector, string, etc) */
    case ___tPROCEDURE:         /* Procedure */
      if ((val >= (___WORD)new_space_bot) && (val < (___WORD)new_space_top))
        {
          printf ("*** ERROR -- Inconsistency detected (pointer to new space during copy)\n");
          printf ("*** Value = 0x%016lx found ", val);
          switch (region)
            {
            case PAIR:
              printf ("in pair object 0x%016lx\n", object);
              break;
            case PLACEHOLDER:
              printf ("in placeholder object 0x%016lx\n", object);
              break;
            case CLOSURE:
              printf ("in closure object 0x%016lx\n", object);
              break;
            case SUBTYPED:
              printf ("in vector-like object 0x%016lx\n", object);
              break;
            default:
              printf ("in ???\n");
              break;
            }
          for (i = 0; i < 20; i++)
            printf ("*** object[%d] = 0x%016lx\n",
                    i, ((___WORD *) (object & -(1 << ___TB)))[i]);
          exit (1);
        }
    }
}

#endif

#ifdef ___MACH_BIG_ENDIAN
#define UPPER 0
#define LOWER 1
#else
#define UPPER 1
#define LOWER 0
#endif

#define ___FWD_FLONUM 0xfff7ffff
#define flonum_get_upper32(x,y) *y = ((___S32*)(x-___tFLONUM))[UPPER]
#define flonum_get_lower32(x,y) *y = ((___S32*)(x-___tFLONUM))[LOWER]
#define flonum_set_upper32(x,y) ((___S32*)(x-___tFLONUM))[UPPER] = y
#define flonum_set_lower32(x,y) ((___S32*)(x-___tFLONUM))[LOWER] = y
#define FLONUMS_PER_FLONUM_CHUNK 256
#define FLONUM_CHUNK_LENGTH \
((((___WS+FLONUMS_PER_FLONUM_CHUNK*8)+(1<<___TB)-1)&-(1<<___TB))-___WS)

void new_flonum (val_ptr)
___WORD *val_ptr;
{
  if (flo_ptr == flo_lim)
  {
    flo_ptr = (((___WORD)bot+___WS+8-1)&-8);
    flo_lim = flo_ptr + FLONUMS_PER_FLONUM_CHUNK*8;
    *bot = ((FLONUM_CHUNK_LENGTH<<___SB)+___sSTRING)<<___TB;
    bot += ((___WS+FLONUM_CHUNK_LENGTH+(1<<___TB)-1)&-(1<<___TB))/___WS;
  }
  *val_ptr = flo_ptr + ___tFLONUM;
  flo_ptr += 8;
}

#define ___FWD_PAIR ((___WORD)(-8<<___TB)+___tSPECIAL)
#define pair_get_car(x,y) *y = ___CAR(x)
#define pair_get_cdr(x,y) *y = ___CDR(x)
#define pair_set_car(x,y) ___CAR(x) = y
#define pair_set_cdr(x,y) ___CDR(x) = y

void new_pair (val_ptr)
___WORD *val_ptr;
{
  top -= 2;
  *val_ptr = (___WORD)top + ___tPAIR;
}

#define ___FWD_PLACEHOLDER ((___WORD)(-8<<___TB)+___tSPECIAL)
#define placeholder_get_car(x,y) *y = ___UNBOX(x,___tPLACEHOLDER)[1]
#define placeholder_get_cdr(x,y) *y = ___UNBOX(x,___tPLACEHOLDER)[0]
#define placeholder_set_car(x,y) ___UNBOX(x,___tPLACEHOLDER)[1] = y
#define placeholder_set_cdr(x,y) ___UNBOX(x,___tPLACEHOLDER)[0] = y

void new_placeholder (val_ptr)
___WORD *val_ptr;
{
  top -= 2;
  *val_ptr = (___WORD)top + ___tPLACEHOLDER;
}

#define subtyped_get_head(x,y) *y = ((___WORD*)(x-___tSUBTYPED))[0]
#define subtyped_set_head(x,y) ((___WORD*)(x-___tSUBTYPED))[0] = y
#define subtyped_subtype(x) ((x>>___TB)&((1<<___SB)-1))
#define is_bvector(x) (x>=___sSTRING)
#define bvector_length(x) (x>>(___SB+___TB))
#define ovector_length(x) (x>>(___SB+___TB))

void new_bvector (val_ptr,length,subtype)
___WORD *val_ptr;
long length;
int subtype;
{
  *val_ptr = (___WORD)bot + ___tSUBTYPED;
  *bot = ((length<<___SB)+subtype)<<___TB;
  bot += ((___WS+length+(1<<___TB)-1)&-(1<<___TB))/___WS;
}

void new_ovector (val_ptr,length,subtype)
___WORD *val_ptr;
long length;
int subtype;
{
  *val_ptr = (___WORD)bot + ___tSUBTYPED;
  *bot = ((length<<___SB)+subtype)<<___TB;
  bot += ((___WS+length*___WS+(1<<___TB)-1)&-(1<<___TB))/___WS;
}

void subtyped_copy_body(x,y,z,ovector)
___WORD x;
___WORD y;
long z;
int ovector;
{
  ___WORD *dst = (___WORD*)(x-___tSUBTYPED+___WS);
  ___WORD *src = (___WORD*)(y-___tSUBTYPED+___WS);
  long n = z;
  while (n>0)
  {
#ifdef SANITY_CHECKS
    if (ovector)
      check_non_new_during_copy (*src,SUBTYPED,y);
#endif
    *dst++ = *src++;
    n--;
  }
}

#define closure_get_head(x,y) *y = ((___WORD*)(x-___tPROCEDURE))[0]
#define closure_set_head(x,y) ((___WORD*)(x-___tPROCEDURE))[0] = y
#define closure_length(x) closure_length2((___WORD*)(x-___tPROCEDURE))
#define closure_prefix_length 2
#define closure_length2(x) (((*(___WORD*)(*(x)-___tPROCEDURE))>>17)&((1<<14)-1))

void new_closure (val_ptr,length)
___WORD *val_ptr;
long length;
{
  *val_ptr = (___WORD)bot + ___tPROCEDURE;
  bot += (((closure_prefix_length+length)*___WS+(1<<___TB)-1)&-(1<<___TB))/___WS;
}

void closure_copy_body(x,y,z)
___WORD x;
___WORD y;
long z;
{
  ___WORD *dst = (___WORD*)(x-___tPROCEDURE);
  ___WORD *src = (___WORD*)(y-___tPROCEDURE);
  long n = closure_prefix_length;
  while (n>0)
  {
    *dst++ = *src++;
    n--;
  }
  n = z;
  while (n>0)
  {
#ifdef SANITY_CHECKS
    check_non_new_during_copy (*src,CLOSURE,y);
#endif
    *dst++ = *src++;
    n--;
  }
}

int in_old_space (val)
___WORD val;
{
  return (val>=(___WORD)old_space_bot) && (val<(___WORD)old_space_top);
}

/*
  The single argument of the procedure "copy_to_new_space" is a pointer to
  a Scheme value.  If the value represents a memory allocated object and
  this object is in "old" space, then the object is copied to "new" space
  and the value is updated so that it refers to the new location of the
  object.  The copy in "old" space is modified to indicate that it has
  been copied and a forwarding pointer to the new location is installed
  in the object.
*/

void copy_to_new_space (orig_ptr)
___WORD *orig_ptr;
{
  ___WORD orig = *orig_ptr;

  switch (___TYP(orig)) /* Dispatch on type of value */
  {

    case ___tFIXNUM: /* Fixed precision integer; immediate */
      break;

    case ___tSPECIAL: /* Special (#f, #t, (), character, etc); immediate */
      break;

    case ___tFLONUM: /* 64 bit floating point number; boxed */

      if (in_old_space (orig))
      {
        ___S32 upper32, lower32;
        flonum_get_upper32 (orig,&upper32); /* get 32 bits with exponent */
        flonum_get_lower32 (orig,&lower32); /* get other 32 bits */
        if (upper32 == ___FWD_FLONUM) /* has it been copied? */
          *orig_ptr = lower32 + (___WORD)new_space_bot;
                      /* forwarding pointer is relative to new space bottom */
        else
        {
          ___WORD copy;
#ifdef DEBUG
          printf ("Copying FLONUM\n");
#endif
          new_flonum (&copy); /* allocate space for new copy */
          flonum_set_upper32 (copy,upper32); /* initialize new copy */
          flonum_set_lower32 (copy,lower32);
          flonum_set_upper32 (orig,___FWD_FLONUM); /* modify old copy */
          flonum_set_lower32 (orig,copy-(___WORD)new_space_bot);
          *orig_ptr = copy;
        }
      }
#ifdef SANITY_CHECKS
      else
        check_non_new (orig);
#endif

      break;

    case ___tPAIR: /* Cons cell */

      if (in_old_space (orig))
      {
        ___WORD car, cdr;
        pair_get_car (orig,&car);
        pair_get_cdr (orig,&cdr);
        if (car == ___FWD_PAIR) /* has it been copied? */
          *orig_ptr = cdr; /* cdr is forwarding pointer */
        else
        {
          ___WORD copy;
#ifdef DEBUG
          printf ("Copying PAIR\n");
#endif
          new_pair (&copy); /* allocate space for new copy */
          pair_set_car (copy,car); /* initialize new copy */
          pair_set_cdr (copy,cdr);
#ifdef SANITY_CHECKS
          check_non_new_during_copy (car,PAIR,orig);
          check_non_new_during_copy (cdr,PAIR,orig);
#endif
          pair_set_car (orig,___FWD_PAIR); /* modify old copy */
          pair_set_cdr (orig,copy);
          *orig_ptr = copy;
        }
      }
#ifdef SANITY_CHECKS
      else
        check_non_new (orig);
#endif

      break;

    case ___tPLACEHOLDER: /* Placeholder */

      if (in_old_space (orig))
      {
        ___WORD car, cdr;
        placeholder_get_car (orig,&car);
        placeholder_get_cdr (orig,&cdr);
        if (car == ___FWD_PLACEHOLDER) /* has it been copied? */
          *orig_ptr = cdr; /* cdr is forwarding pointer */
        else
        {
          ___WORD copy;
#ifdef DEBUG
          printf ("Copying PLACEHOLDER\n");
#endif
          new_placeholder (&copy); /* allocate space for new copy */
          placeholder_set_car (copy,car); /* initialize new copy */
          placeholder_set_cdr (copy,cdr);
#ifdef SANITY_CHECKS
          check_non_new_during_copy (car,PLACEHOLDER,orig);
          check_non_new_during_copy (cdr,PLACEHOLDER,orig);
#endif
          placeholder_set_car (orig,___FWD_PLACEHOLDER); /* modify old copy */
          placeholder_set_cdr (orig,copy);
          *orig_ptr = copy;
        }
      }
#ifdef SANITY_CHECKS
      else
        check_non_new (orig);
#endif

      break;

    case ___tSUBTYPED: /* Object with subtype (vector, string, etc) */

      if (in_old_space (orig))
      {
        ___WORD head;
        subtyped_get_head (orig,&head);
        if (___TYP(head) == ___tSUBTYPED) /* has it been copied? */
          *orig_ptr = head; /* head is forwarding pointer */
#ifdef SANITY_CHECKS
        else if (___TYP(head) != 0)
          fatal_error ("Corrupt head\n");
#endif
        else
        {
          ___WORD copy;
          long len;
          int subtype = subtyped_subtype (head); /* get subtype */
          if (is_bvector (subtype)) /* is it a byte vector? (e.g. a string) */
          {
            len = bvector_length (head); /* get length in bytes */
#ifdef DEBUG
            printf ("Copying BYTE-VECTOR of length %ld\n",len);
#endif
            new_bvector (&copy,len,subtype); /* allocate space for new copy */
            len = (len+___WS-1)/___WS; /* convert length to words */
            subtyped_copy_body (copy,orig,len,0); /* initialize new copy */
          }
          else /* it must be an object vector (e.g. a vector) */
          {
            len = ovector_length (head); /* get length in words */
#ifdef DEBUG
            printf ("Copying OBJECT-VECTOR of length %ld\n",len);
#endif
            new_ovector (&copy,len,subtype); /* allocate space for new copy */
            subtyped_copy_body (copy,orig,len,1); /* initialize new copy */
          }
          subtyped_set_head (orig,copy); /* modify old copy */
          *orig_ptr = copy;
        }
      }
#ifdef SANITY_CHECKS
      else
        check_non_new (orig);
#endif

      break;

    case ___tPROCEDURE: /* Procedure */

      if (in_old_space (orig))
      {

        /* Since this procedure is in the heap, then it can only
           be a closure.  Constant procedures are never in the heap. */

        ___WORD head;
        closure_get_head (orig,&head);
        if (___TYP(head) == 0) /* has it been copied? */
          *orig_ptr = ___tPROCEDURE + head; /* head is forwarding pointer */
#ifdef SANITY_CHECKS
        else if (___TYP(head) != ___tPROCEDURE)
          fatal_error ("Corrupt head\n");
#endif
        else
        {
          ___WORD copy;
          long len = closure_length (orig); /* get number of variables */
#ifdef DEBUG
          printf ("Copying CLOSURE with %ld variables\n",len);
#endif
          new_closure (&copy,len); /* allocate space for new copy */
          closure_copy_body (copy,orig,len); /* initialize new copy */
          closure_set_head (orig,copy-___tPROCEDURE); /* update old copy */
          *orig_ptr = copy;
        }
      }
#ifdef SANITY_CHECKS
      else
        check_non_new (orig);
#endif

      break;

    default:
      printf ("*** ERROR -- Value with unknown tag encountered\n");
      printf ("*** Object = 0x%016lx at address 0x%016lx\n",orig,orig_ptr);
      printf ("*** old_space_bot = 0x%016lx\n",old_space_bot);
      printf ("*** old_space_top = 0x%016lx\n",old_space_top);
      printf ("*** new_space_bot = 0x%016lx\n",new_space_bot);
      printf ("*** new_space_top = 0x%016lx\n",new_space_top);
      exit(1);
  }
}

extern ___WORD ___internal_return;

void copy_frame_to_new_space (orig_ptr, ra)
___WORD *orig_ptr;
___WORD ra;
{
  ___WORD frame;

next_frame:

  frame = *orig_ptr;

  if (frame != 0)
  if (___TYP(frame) == ___tSUBTYPED)
    copy_to_new_space (orig_ptr);
  else
    {
      /* continuation's frame is in the stack */

      ___WORD *fp, frame_ra;
      int fs, link, i;

      if (ra == ___internal_return)
        {
          ___WORD fs_link = ((___WORD *)(((___WORD*)frame)[0]-___tPROCEDURE))[0];
          fs = (fs_link>>3)&((1<<14)-1);
          link = (fs_link>>17)&((1<<14)-1);
        }
      else
        {
          ___WORD fs_link = ((___WORD *)(ra-___tPROCEDURE))[0];
          fs = (fs_link>>3)&((1<<14)-1);
          link = (fs_link>>17)&((1<<14)-1);
        }
      fp = ((___WORD*)frame)+fs;
      frame_ra = fp[-link-1];

      if (___TYP(frame_ra) == ___tSUBTYPED) /* has it been copied? */
        *orig_ptr = frame_ra;
      else
        {
          ___WORD copy, *f;
          long len = fs+1;
          new_ovector (&copy,len,___sFRAME); /* allocate space for new copy */

          f = len+(___WORD*)(copy-___tSUBTYPED+___WS);

          if (frame_ra == ps->handler_break)
          {
            /* first frame of that section */

            f[-fs-1] = fp[1];
            ra       = fp[0];
            for (i=-fs; i<0; i++)
              f[i] = fp[i];

            f[-link-1] = ra;
          }
        else
          {
            /* not the first frame of that section */

            f[-fs-1] = (___WORD)fp;
            ra       = frame_ra;

            for (i=-fs; i<0; i++)
              f[i] = fp[i];
          }

        fp[-link-1] = copy; /* forwarding pointer */
        *orig_ptr = copy;

        orig_ptr = &f[-fs-1];
        goto next_frame;
        }
    }

}

/*---------------------------------------------------------------------------*/

void print_state ()
{
  printf("\n---------------------------------------------------------------\n");
  printf("*** r0 = 0x%016lx\n",ps->r[0]);
  printf("*** r1 = 0x%016lx\n",ps->r[1]);
  printf("*** r2 = 0x%016lx\n",ps->r[2]);
  printf("*** r3 = 0x%016lx\n",ps->r[3]);
  printf("*** r4 = 0x%016lx\n",ps->r[4]);
  printf("*** sba= 0x%016lx\n",ps->stack_base);
  printf("*** sb = 0x%016lx\n",ps->sb);
  printf("*** sp = 0x%016lx\n",ps->sp);
  printf("*** sl = 0x%016lx\n",ps->sl);
  printf("*** stk= 0x%016lx\n",ps->stack);
  printf("*** hp = 0x%016lx\n",ps->hp);
  printf("*** hl = 0x%016lx\n",ps->hl);
  printf("*** h  = 0x%016lx\n",ps->heap);
  printf("*** temp1 = 0x%016lx\n",ps->temp1);
  printf("\n");
  printf("*** old_space_bot = 0x%016lx\n",old_space_bot);
  printf("*** old_space_top = 0x%016lx\n",old_space_top);
  printf("*** new_space_bot = 0x%016lx\n",new_space_bot);
  printf("*** new_space_top = 0x%016lx\n",new_space_top);
}

void scan_closure_and_copy_subobjects (head,ptr,follow)
___WORD head;
___WORD *ptr;
___WORD **follow;
{
  long len = closure_length2(ptr-1); /* get number of variables */
#ifdef DEBUG
  printf ("Scanning CLOSURE with %ld variables\n",len);
#endif
#ifdef SANITY_CHECKS
  scanned_region = CLOSURE;
  scanned_object = (___WORD)ptr - ___WS + ___tPROCEDURE;
#endif
  ptr += closure_prefix_length-1;
  while (len--)
  {
    copy_to_new_space (ptr);
    ptr++;
  }
  *follow = ptr;
}

void scan_subtyped_and_copy_subobjects (head,ptr,follow)
___WORD head;
___WORD *ptr;
___WORD **follow;
{
  int subtype = subtyped_subtype (head); /* get subtype */
  if (is_bvector (subtype)) /* is it a byte vector? (e.g. a string) */
  {
    long len = bvector_length (head); /* get length in bytes */
#ifdef DEBUG
    printf ("Scanning BYTE-VECTOR of length %ld\n",len);
#endif
    ptr += (len+___WS-1)/___WS; /* skip body */
  }
  else /* it must be an object vector (e.g. a vector) */
  {
    long len = ovector_length (head); /* get length in words */
#ifdef DEBUG
    printf ("Scanning OBJECT-VECTOR of length %ld\n",len);
#endif
#ifdef SANITY_CHECKS
    scanned_region = SUBTYPED;
    scanned_object = (___WORD)ptr - ___WS + ___tSUBTYPED;
#endif
    if (subtype == ___sSYMBOL) /* don't scan pointer to global var */
    {
      len--;
      while (len--)
      {
        copy_to_new_space (ptr);
        ptr++;
      }
      ptr++;
    }
    else if (subtype == ___sFRAME) /* scan link only if in old space */
    {
      if (in_old_space (ptr[0]))
        copy_to_new_space (ptr);
      len--;
      ptr++;
      while (len--)
      {
        copy_to_new_space (ptr);
        ptr++;
      }
    }
    else if (subtype == ___sCONTFRAME) /* scan link specially */
    {
      copy_frame_to_new_space (ptr, ptr[1]);
      len--;
      ptr++;
      while (len--)
      {
        copy_to_new_space (ptr);
        ptr++;
      }
    }
    else
      while (len--)
      {
        copy_to_new_space (ptr);
        ptr++;
      }
  }
  *follow = ptr;
}

void scan_new_space_and_copy_subobjects ()
{
  ___WORD *scan_bot = new_space_bot;
  ___WORD *scan_top = new_space_top;

  loop:

#ifdef SANITY_CHECKS
  scanned_region = PAIR;
#endif

  while (scan_top>top)
  {
    scan_top--;
    copy_to_new_space (scan_top);
  }

  if (scan_bot<bot)
  {
    ___WORD head = *scan_bot++;

    if (___TYP(head) == ___tPROCEDURE)
      scan_closure_and_copy_subobjects (head,scan_bot,&scan_bot);
#ifdef SANITY_CHECKS
    else if (___TYP(head) != 0)
      fatal_error ("Corrupt head\n");
#endif
    else
      scan_subtyped_and_copy_subobjects (head,scan_bot,&scan_bot);

    scan_bot = (___WORD*)(((___WORD)scan_bot+(1<<___TB)-1)&-(1<<___TB));

    goto loop;
  }
}

void copy_root_set_to_new_space ()
{
  long i;

  /* update constant area */

#ifdef SANITY_CHECKS
  scanned_region = CONSTANT;
#endif

  i = 0;
  while (___module_tbl[i] != 0)
  {
    int j;
    ___WORD *cns = ___ALIGNUP((___module_tbl[i]->cns_tbl));
    ___WORD *sub_tbl = (___WORD*)___module_tbl[i]->sub_tbl;

    for (j=___module_tbl[i]->cns_count*2-1; j>=0; j--)
      copy_to_new_space (cns+j);

    for (j=___module_tbl[i]->sub_count-1; j>=0; j--)
    {
      ___WORD *ptr = (___WORD*)(sub_tbl[j]&-(1<<___TB));
      ___WORD head = *ptr++;
      scan_subtyped_and_copy_subobjects (head,ptr,&ptr);
    }

    if (___module_tbl[i]->sym_names != 0)
      for (j=___module_tbl[i]->sym_count-1; j>=0; j--)
        copy_to_new_space (___module_tbl[i]->sym_tbl+j);

    i++;
  }

  /* update registers */

#ifdef SANITY_CHECKS
  scanned_region = REGISTER;
#endif

  for (i=___MAX_NB_GVM_REGS-1; i>=0; i--)
    copy_to_new_space (ps->r+i);

  /* update global variables */

#ifdef SANITY_CHECKS
  scanned_region = GLOBAL_VAR;
#endif

  {
    ___glo_struct *p = ps->glo_list;

    while (p != 0)
    {
      copy_to_new_space (&p->value);
      p = p->next;
    }
  }

  /* update stack */

#ifdef SANITY_CHECKS
  scanned_region = STACK;
#endif

  copy_frame_to_new_space (&ps->sb[1], ps->sb[0]);

  for (i=ps->sb-ps->sp-1; i>=0; i--)
    copy_to_new_space (ps->sp+i);
}

void move_stack ()
{
  ___WORD *sp = ps->sp;
  ___WORD *p1 = ps->sb+2;
  ___WORD *p2 = ps->stack_base;
  if (p1 != p2)
    {
      ps->sb = p2-2;
      while (p1 != sp)
        *--p2 = *--p1;
      ps->sp = p2;
    }
}

void gc (___pstate,words)
___processor ___pstate;
long words;
{
  ___WORD *semispace1_bot, *semispace2_bot;

  if ((___G__23_23gc_2dreport.value != ___FAL) || debug_level)
  {
    printf ("*** GC started...");
    fflush (stdout);
  }

  semispace1_bot = ___pstate->heap;
  semispace2_bot = semispace1_bot+semispace_length;

  if (___pstate->hl<semispace2_bot)
  {
    old_space_bot = semispace1_bot;
    new_space_bot = semispace2_bot;
  }
  else
  {
    old_space_bot = semispace2_bot;
    new_space_bot = semispace1_bot;
  }

  old_space_top = old_space_bot+semispace_length;
  new_space_top = new_space_bot+semispace_length;

  bot = new_space_bot; /* bottom allocation pointer; for objects with a head */
  top = new_space_top; /* top allocation pointer; for objects without a head */

  flo_ptr = 0; /* flonum allocation pointers */
  flo_lim = 0;

  ps = ___pstate;

  if (debug_level)
    print_state ();

  copy_root_set_to_new_space ();

  scan_new_space_and_copy_subobjects ();

  ___pstate->hp = top;
  ___pstate->hl = bot+semispace_fudge;

  move_stack ();

  if (debug_level)
    print_state ();

  if ((___G__23_23gc_2dreport.value != ___FAL) || debug_level)
  {
    printf ("done (%d%% of heap in use)\n",
            100*(semispace_length-(top-bot))/semispace_length);
    fflush (stdout);
  }

  if (___pstate->hp-___pstate->hl<words)
  {
    printf("*** ERROR -- Heap overflow\n");
    exit(1);
  }
}
