/* file: "os.c" */

/* Copyright (C) 1994, 1995 by Marc Feeley, All Rights Reserved. */

/*
This module implements the operating system specific routines including:

  - dynamic linking
  - user and timer interrupt handling
  - elapsed time
*/

#define ___VERSION 4
#include "gambit.h"
___BEGIN_MODULE

/*---------------------------------------------------------------------------*/

/* For dynamic linking */

#ifdef sparc
#include <dlfcn.h>
#endif

#ifdef sgi
#include <dlfcn.h>
#endif

#ifdef __alpha
#include <dlfcn.h>
#endif

#ifdef hpux
#include <dl.h>
#endif

#ifdef _X86_
#include <windows.h>
#endif

void *os_dyn_link (path)
char *path;
{
  char *module_linker = "___init_module";

#ifdef sparc
#ifdef RTLD_NOW
  void *descr = dlopen (path, RTLD_NOW);
#else
  void *descr = dlopen (path, 1);
#endif
  if (descr == 0) return 0;
  return dlsym (descr, module_linker);
#endif

#ifdef sgi
  void *descr = dlopen (path, RTLD_NOW);
  if (descr == 0) return 0;
  return dlsym (descr, module_linker);
#endif

#ifdef __alpha
  void *descr = dlopen (path, RTLD_NOW);
  if (descr == 0) return 0;
  return dlsym (descr, module_linker);
#endif

#ifdef hpux
  void *f;
  shl_t descr = shl_load (path, BIND_IMMEDIATE, 0);
  if (descr == 0) return 0;
  if (shl_findsym (&descr, module_linker, TYPE_PROCEDURE, &f)) return 0;
  return f;
#endif

#ifdef _X86_
  HANDLE descr = LoadLibrary (path);
  if (descr < 32) return 0;
  return GetProcAddress (descr, module_linker);
#endif
}

/*---------------------------------------------------------------------------*/

#include "setup.h"
#include "gc.h"

#include <time.h>

#ifndef WIN32
#ifndef __MWERKS__
#include <sys/time.h>
#endif
#endif

#ifdef CLOCKS_PER_SEC
static clock_t clock_start;
#else
#include <sys/times.h>
#include <sys/unistd.h>
#endif

/*---------------------------------------------------------------------------*/

void raise_interrupt( n )
int n;
{
  pstate->intr_flag[n] = 1;
  if (pstate->intr_enabled)
    pstate->sl = pstate->stack_base;
}

/*---------------------------------------------------------------------------*/

/* OS specific stuff */

#ifdef GO32

#include <dpmi.h>
#include <go32.h>
#include <dos.h>

static void user_interrupt_handler()
{
  raise_interrupt(0);
}

static void timer_interrupt_handler()
{
  raise_interrupt(1);
}

static _go32_dpmi_registers regs;
static _go32_dpmi_seginfo old_vector, info;

static void install_interrupt_handlers()
{
  _go32_dpmi_get_real_mode_interrupt_vector(0x1b, &old_vector);
  info.pm_offset = (int)user_interrupt_handler;
  _go32_dpmi_allocate_real_mode_callback_iret(&info, &regs);
  _go32_dpmi_set_real_mode_interrupt_vector(0x1b, &info);
  _go32_want_ctrl_break(1);
  setcbrk(1);
     
  _go32_dpmi_get_protected_mode_interrupt_vector(8, &old_vector);
  info.pm_offset = (int)timer_interrupt_handler;
  info.pm_selector = _go32_my_cs();
  _go32_dpmi_chain_protected_mode_interrupt_vector(8, &info);
}

int os_shell_command( cmd )
char *cmd;
{
  return system( cmd );
}

#else

#ifdef __MWERKS__

static void install_interrupt_handlers()
{
}

int os_shell_command( cmd )
char *cmd;
{
}

#else

#include <signal.h>

#ifdef MACHTEN

static void user_interrupt_handler(sig, code, scp)
int sig, code;
struct sigcontext *scp;
{
  raise_interrupt(0);
}

static void timer_interrupt_handler(sig, code, scp)
int sig, code;
struct sigcontext *scp;
{
  raise_interrupt(1);
}

void set_timer_interval( interval )
int interval;
{
  struct itimerval virtual_tv;
  int secs = interval/1000;
  int usecs = (interval%1000)*1000;
  virtual_tv.it_interval.tv_sec  = secs;
  virtual_tv.it_interval.tv_usec = usecs;
  virtual_tv.it_value.tv_sec     = secs;
  virtual_tv.it_value.tv_usec    = usecs;
  setitimer( ITIMER_VIRTUAL, &virtual_tv, 0 );
}

static void install_interrupt_handlers()
{
  struct sigvec vec;

  vec.sv_handler = user_interrupt_handler;
  vec.sv_flags = SV_INTERRUPT;
  vec.sv_mask = -1;
  sigvec(SIGINT, &vec, 0);

  vec.sv_handler = timer_interrupt_handler;
  vec.sv_flags = SV_INTERRUPT;
  vec.sv_mask = -1;
  sigvec(SIGVTALRM, &vec, 0);

  set_timer_interval( 100 );
}

#else

#ifdef WIN32

static void user_interrupt_handler (int interrupt)
{
  raise_interrupt(0);
}

static void install_interrupt_handlers()
{
  signal (SIGINT, user_interrupt_handler);
}

#else

static void user_interrupt_handler()
{
  raise_interrupt(0);
}

static void timer_interrupt_handler()
{
  raise_interrupt(1);
}

void set_timer_interval( interval )
int interval;
{
  struct itimerval virtual_tv;
  int secs = interval/1000;
  int usecs = (interval%1000)*1000;
  virtual_tv.it_interval.tv_sec  = secs;
  virtual_tv.it_interval.tv_usec = usecs;
  virtual_tv.it_value.tv_sec     = secs;
  virtual_tv.it_value.tv_usec    = usecs;
  setitimer( ITIMER_VIRTUAL, &virtual_tv, 0 );
}

static void install_interrupt_handlers()
{
  struct sigaction act;

  act.sa_handler = user_interrupt_handler;
  act.sa_flags = 0;
#ifdef SA_INTERRUPT
  act.sa_flags |= SA_INTERRUPT;
#endif
  sigemptyset(&act.sa_mask);
  sigaction(SIGINT, &act, 0);

  act.sa_handler = timer_interrupt_handler;
  act.sa_flags = 0;
#ifdef SA_INTERRUPT
  act.sa_flags |= SA_INTERRUPT;
#endif
  sigemptyset(&act.sa_mask);
  sigaction(SIGVTALRM, &act, 0);

  set_timer_interval( 100 );
}

#endif
#endif

int os_shell_command( cmd )
char *cmd;
{
  int result;
  set_timer_interval( 0 );
  result = system( cmd );
  set_timer_interval( 100 );
  return result;
}

#endif
#endif

unsigned long os_clock()
{
  unsigned long msec;

#ifdef CLOCKS_PER_SEC
  return (clock()-clock_start)*1000.0/CLOCKS_PER_SEC;
#else
  struct tms ts;
  unsigned long ticks_per_sec = sysconf(_SC_CLK_TCK);
  times(&ts);
  return ts.tms_utime*1000.0/ticks_per_sec;
#endif
}

void os_init()
{
  install_interrupt_handlers();

#ifdef CLOCKS_PER_SEC
  clock_start = clock();
#endif
}
