/* file: "setup.c" */

/* Copyright (C) 1994, 1995 by Marc Feeley, All Rights Reserved. */

/* This module contains the setup routines for the Gambit-C system */

#define ___VERSION 4
#include "gambit.h"
___BEGIN_MODULE

___DEMAND_GLO(___G__23_23symbol_2dtable)
___DEMAND_GLO(___G__23_23exec_2dvector)
___DEMAND_GLO(___G__23_23argv)
___DEMAND_GLO(___G__23_23datadir)

#include <stdio.h>
#include <stdlib.h>

#include "setup.h"
#include "os.h"

___module ___module_tbl[100];
int nb_modules;

int debug_level = 0; /* non-zero if debugging information is requested    */

___processor pstate; /* processor state (e.g. registers, allocation ptrs) */

long semispace_length; /* length of a semispace in words              */
long semispace_fudge;  /* length of the semispace fudge area in words */
long stack_length;     /* length of the stack in words                */
long stack_fudge;      /* length of the stack fudge area in words     */

/*---------------------------------------------------------------------------*/

/* Memory allocation parameters */

#define K 1024

#define DEFAULT_STACK_LENGTH_IN_K 100
#define MIN_STACK_LENGTH_IN_K     20
#define STACK_FUDGE_IN_K          10

#define DEFAULT_HEAP_LENGTH_IN_K  8192
#define MIN_HEAP_LENGTH_IN_K      100
#define HEAP_FUDGE_IN_K           10

#define SYMBOL_TABLE_LENGTH 43

void *malloc8(n) /* allocate n bytes, aligned to an 8 byte multiple */
long n;
{
  void *p = malloc(n+7);
  if (p==NULL)
  {
    fprintf( stderr, "*** ERROR -- Insufficient memory\n" );
    exit(1);
  }
  return (void*)(-8&(7+(long)p));
}

___WORD make_ovector(n,subtype) /* allocate an object vector from heap */
int n;
int subtype;
{
  ___WORD result;
  long words;
  words = ((n*___WS+___WS+(1<<___TB)-1)&-(1<<___TB))/___WS;
  if (pstate->hp-pstate->hl<words)
    {
      fprintf( stderr, "*** ERROR -- Heap overflow during setup\n" );
      exit(1);
    }
  pstate->hp -= words;
  result = (___WORD)pstate->hp + ___tSUBTYPED;
  ___HEADER(result)=(n<<___SB+___TB)+(subtype<<___TB);
  return result;
}

___WORD make_pair(car,cdr) /* allocate a pair from heap */
___WORD car, cdr;
{
  ___WORD result;
  if (pstate->hp-pstate->hl<2)
    {
      fprintf( stderr, "*** ERROR -- Heap overflow during setup\n" );
      exit(1);
    }
  pstate->hp -= 2;
  result = (___WORD)pstate->hp + ___tPAIR;
  ___SETCAR(result,car)
  ___SETCDR(result,cdr)
  return result;
}

___WORD c_string_to_scheme_string(s) /* convert a string from C to Scheme */
char *s;
{
  ___WORD result;
  long words;
  int n;
  n = 0;
  while (s[n] != '\0') n++;
  words = ((n+1+___WS+(1<<___TB)-1)&-(1<<___TB))/___WS;
  if (pstate->hp-pstate->hl<words)
    {
      fprintf( stderr, "*** ERROR -- Heap overflow during setup\n" );
      exit(1);
    }
  pstate->hp -= words;
  result = (___WORD)pstate->hp + ___tSUBTYPED;
  ___HEADER(result)=(n+1<<___SB+___TB)+(___sSTRING<<___TB);
  while (n >= 0)
    {
      ___STRINGSET(result,___FIX(n),___CHR(s[n]))
      n--;
    }
  return result;
}

long hash_string(s)
char *s;
{
  long h = 0;
  char *p = s;
  unsigned char c;
  while ((c=*p++) != '\0')
    h = (long)((h<<8)+c) % SYMBOL_TABLE_LENGTH;
  return h;
}

___WORD make_symbol(s)
char *s;
{
  long h = hash_string(s);
  ___WORD probe = ___VECTORREF(___G__23_23symbol_2dtable.value,___FIX(h));
  ___WORD sym;

  while (probe != ___NUL)
    {
      char *p1, *p2;
      sym = ___CAR(probe);
      p1 = s;
      p2 = (char*)(___VECTORREF(sym,___FIX(0))-___tSUBTYPED+___WS);
      while ((*p1!='\0')&&(*p2!='\0')&&(*p1==*p2)) { p1++; p2++; }
      if (*p1==*p2)
	return sym;
      else
	probe = ___CDR(probe);
    }

  sym = make_ovector(3,___sSYMBOL);
	
  ___VECTORSET(sym,___FIX(0),c_string_to_scheme_string(s))
  ___VECTORSET(sym,___FIX(1),___FIX(h))
  ___VECTORSET(sym,___FIX(2),0)

  ___VECTORSET(___G__23_23symbol_2dtable.value,
	       ___FIX(h),
	       make_pair(sym,___VECTORREF(___G__23_23symbol_2dtable.value,___FIX(h))))

  return sym;
}

___glo_struct *make_global(s)
char *s;
{
  ___WORD sym = make_symbol(s);
  ___WORD glo = ___VECTORREF(sym,___FIX(2));
  if (glo != 0)
    return (___glo_struct*)glo;
  else
    {
      ___glo_struct *p = (___glo_struct*)malloc(sizeof(___glo_struct));
      if (p == 0)
	{
	  fprintf( stderr, "*** ERROR -- Can't allocate global var\n" );
	  exit(1);
	}
      p->value = ___UNB;
      p->next = pstate->glo_list;
      pstate->glo_list = p;
      ___VECTORSET(sym,___FIX(2),(___WORD)p)
      return p;
    }
}

/*---------------------------------------------------------------------------*/

/* Alignment of objects */

___WORD *align( from, len )
___WORD *from;
int len;
{
  ___WORD *to = ___ALIGNUP(from);
  if (from != to)
    {
      /* move object up */
      int i;
      for (i=len-1; i>=0; i--) to[i] = from[i];
      from[0] = 0;
    }
  return to;
}

___WORD align_subtyped( ptr )
___WORD *ptr;
{
  ___WORD header = ptr[0];
  int subtype = (header&(1<<___SB+___TB)-1)>>___TB;
  int len = header>>___SB+___TB;
  int wlen;
  switch (subtype)
    {
    case ___sSYMBOL:
    case ___sVECTOR:
    case ___sRATNUM:
    case ___sCPXNUM:
      wlen = len+1;
      break;
    case ___sSTRING:
    case ___sBIGNUM:
      wlen = (len+___WS-1)/___WS+1;
      break;
    default:
      fprintf( stderr, "*** ERROR -- Unknown subtype during fixup\n" );
      exit(1);
    }
  return ___tSUBTYPED+(___WORD)align( ptr, wlen );
}

/*---------------------------------------------------------------------------*/

/* Routines to setup a module for execution */

void fixref( p, sym_tbl, cns_tbl, flo_tbl, sub_tbl )
___WORD *p;
___WORD *sym_tbl, *cns_tbl, *flo_tbl, *sub_tbl;
{
  ___WORD v = *p;
  switch (___TYP(v))
    {
    case ___rSYM: /*SYM*/
      *p = sym_tbl[___INT(v)];
      break;

    case ___rCNS: /*CNS*/
      *p = ___BOX(___ALIGNUP((&cns_tbl[2*___INT(v)])),___tPAIR);
      break;

    case ___rFLO: /*FLO*/
      *p = ___BOX(___ALIGNUP((&flo_tbl[___FW*___INT(v)])),___tFLONUM);
      break;

    case ___rSUB: /*SUB*/
      *p = sub_tbl[___INT(v)];
      break;
    }
}

void setup_module(module)
___module module;
{
  int i, j;
  ___WORD *cns, *flo;

  char *name       = module->name;
  ___glo_struct **glo_tbl= module->glo_tbl;
  int glo_count    = module->glo_count;
  char **glo_names = module->glo_names;
  ___WORD *sym_tbl = (___WORD*)module->sym_tbl;
  int sym_count    = module->sym_count;
  char **sym_names = module->sym_names;
  ___WORD *lp      = module->lp;
  ___WORD *lbl_tbl = (___WORD*)module->lbl_tbl;
  int lbl_count    = module->lbl_count;
  ___WORD *cns_tbl = module->cns_tbl;
  int cns_count    = module->cns_count;
  ___WORD *flo_tbl = module->flo_tbl;
  int flo_count    = module->flo_count;
  ___WORD *sub_tbl = (___WORD*)module->sub_tbl;
  int sub_count    = module->sub_count;

  if (debug_level)
    fprintf( stderr, "setting up module \"%s\"\n", name );

  if (lbl_count > 0)
    {
      ___WORD *new;
      new = align( lbl_tbl, lbl_count*2 );
      *lp = (___WORD)new + ___tPROCEDURE;
    }

  cns = align( cns_tbl, 2*cns_count );

  if (glo_names != 0)
    {
      i = 0;
      while (glo_names[i] != 0)
	{
	  glo_tbl[i] = make_global(glo_names[i]);
	  i++;
	}
    }

  if (sym_names != 0)
    {
      i = 0;
      while (sym_names[i] != 0)
	{
	  sym_tbl[i] = make_symbol(sym_names[i]);
	  i++;
	}
    }
  else
    for (i=sym_count-1; i>=0; i--)
      sym_tbl[i] = ___tSUBTYPED+(___WORD)___ALIGNUP( (___WORD*)sym_tbl[i] );

  flo = align( flo_tbl, ___FW*flo_count );

  for (i=sub_count-1; i>=0; i--)
    sub_tbl[i] = align_subtyped( (___WORD*)sub_tbl[i] );

  for (i=cns_count*2-1; i>=0; i--)
    fixref( cns+i, sym_tbl, cns_tbl, flo_tbl, sub_tbl );

  for (j=sub_count-1; j>=0; j--)
    {
      ___WORD *p = (___WORD*)(sub_tbl[j]-___tSUBTYPED);
      ___WORD header = p[0];
      int subtype = (header&(1<<___SB+___TB)-1)>>___TB;
      int len = header>>___SB+___TB;
      switch (subtype)
	{
	case ___sSYMBOL:
	case ___sVECTOR:
	case ___sRATNUM:
	case ___sCPXNUM:
	  for (i=1; i<=len; i++)
	    fixref( p+i, sym_tbl, cns_tbl, flo_tbl, sub_tbl );
	}
    }

  module->init_proc();
}

char hex_tbl[] = "0123456789abcdef";

int load_object_file(s)
char *s;
{
  void *f = os_dyn_link (s);
  if (f == 0) return 1;
  ___module_tbl[nb_modules] = ((___module(*)())f)();
  setup_module( ___module_tbl[nb_modules] );
  nb_modules++;
  return 0;
}

/*---------------------------------------------------------------------------*/

/* System setup */

void gambit_setup( argc, argv, stack_len_in_k, heap_len_in_k )
int argc;
char **argv;
int stack_len_in_k, heap_len_in_k;
{
  int stack_len, heap_len;
  int i, n;
  ___WORD symbol_table;

  /* compute size of stack and heap */

  if (stack_len_in_k > 0)
    stack_len = stack_len_in_k*K;
  else
    stack_len = DEFAULT_STACK_LENGTH_IN_K*K;

  if (heap_len_in_k > 0)
    heap_len = heap_len_in_k*K;
  else
    heap_len = DEFAULT_HEAP_LENGTH_IN_K*K;

  /* check for minimum stack and heap sizes */

  if (stack_len < MIN_STACK_LENGTH_IN_K*K)
    {
      stack_len = MIN_STACK_LENGTH_IN_K*K;
      fprintf( stderr,
	       "*** WARNING -- Minimum size stack (%dK) is being allocated\n",
	       MIN_STACK_LENGTH_IN_K );
    }

  if (heap_len < MIN_HEAP_LENGTH_IN_K*K)
    {
      heap_len = MIN_HEAP_LENGTH_IN_K*K;
      fprintf( stderr,
	       "*** WARNING -- Minimum size heap (%dK) is being allocated\n",
	       MIN_HEAP_LENGTH_IN_K );
    }

  /* allocate stack and heap */

  semispace_length = heap_len/___WS/2;
  semispace_fudge  = HEAP_FUDGE_IN_K*K/___WS;
  stack_length     = stack_len/___WS;
  stack_fudge      = STACK_FUDGE_IN_K*K/___WS;

  pstate = (___processor)malloc8( sizeof(___processor_struct) );
  pstate->heap  = (___WORD*)malloc8( semispace_length*2*___WS );
  pstate->stack = (___WORD*)malloc8( stack_length*___WS );

  /* setup stack and heap allocation pointers */

  pstate->hp = pstate->heap+semispace_length;
  pstate->hl = pstate->heap+semispace_fudge;
  pstate->stack_base = pstate->stack+stack_length;
  pstate->sp = pstate->stack_base;

  /* setup interrupt system */

  pstate->sl = pstate->stack+stack_fudge;

  pstate->intr_enabled = 1;
  pstate->intr_flag[0] = 0;
  pstate->intr_flag[1] = 0;

  /* create and initialize global variables and primitives */

  {
    ___WORD **p = ___sym_list;

    while (p != 0)
      {
        ___WORD *sym_ptr;
        ___glo_struct *glo;

        sym_ptr = (___WORD*)p;

        p = (___WORD**)sym_ptr[0];
        glo = (___glo_struct*)sym_ptr[3];

        sym_ptr[2] = glo->value;

        glo->value = ___UNB;
      }
    }

  /* create symbol table */

  symbol_table = make_ovector(SYMBOL_TABLE_LENGTH,___sVECTOR);

  for (i=0; i<SYMBOL_TABLE_LENGTH; i++)
    ___VECTORSET(symbol_table,___FIX(i),___NUL)

  {
    ___WORD **p = ___sym_list;
    pstate->glo_list = 0;

    while (p != 0)
    {
      long h;
      ___WORD str, sym;
      ___WORD *sym_ptr;
      ___glo_struct *glo;

      sym_ptr = (___WORD*)p;

      p = (___WORD**)sym_ptr[0];
      str = align_subtyped( (___WORD*)sym_ptr[1] );
      glo = (___glo_struct*)sym_ptr[3];

      glo->next = pstate->glo_list;
      pstate->glo_list = glo;

      *sym_ptr = (3<<___SB+___TB)+(___sSYMBOL<<___TB);

      sym = align_subtyped( sym_ptr );

      h = hash_string((char*)(str-___tSUBTYPED+___WS));

      ___VECTORSET(sym,___FIX(0),str)
      ___VECTORSET(sym,___FIX(1),___FIX(h))
      ___VECTORSET(sym,___FIX(2),(___WORD)glo)

      ___VECTORSET(symbol_table,
		   ___FIX(h),
		   make_pair(sym,___VECTORREF(symbol_table,___FIX(h))))
    }
  }

  ___G__23_23symbol_2dtable.value = symbol_table;

  /* setup each module */

  n = 0;
  nb_modules = 0;
  while (___modules[nb_modules] != 0)
    {
      ___module_tbl[nb_modules] = ((___module(*)())___modules[nb_modules])();
      setup_module( ___module_tbl[nb_modules] );
      if (___module_tbl[nb_modules]->lbl_count != 0) n++;
      nb_modules++;
    }

  /* create execution vector */

  ___G__23_23exec_2dvector.value = make_ovector(n,___sVECTOR);

  n = 0;
  for (i=0; i<nb_modules; i++)
    if (___module_tbl[i]->lbl_count != 0)
      {
	___VECTORSET(___G__23_23exec_2dvector.value,___FIX(n),
		     *___module_tbl[i]->lp+2*___WS)
	n++;
      }

  /* create argument vector */

  ___G__23_23argv.value = make_ovector(argc,___sVECTOR);

  ___VECTORSET(___G__23_23argv.value,___FIX(0),
               c_string_to_scheme_string(argv[0]))

  for (i=1; i<argc; i++)
    ___VECTORSET(___G__23_23argv.value,___FIX(i),
                 c_string_to_scheme_string(argv[i]))

  /* create ##datadir string */

#ifdef DATADIR
  ___G__23_23datadir.value = c_string_to_scheme_string(DATADIR);
#else
  ___G__23_23datadir.value = ___FAL;
#endif

  /* setup call to kernel to start executing program */

  pstate->sp[-1] = 0; /* create "break frame" of stack's initial top section */
  pstate->sp[-2] = 0;
  pstate->sp -= 2;
  pstate->sb = pstate->sp;

  for (i=1; i<___MAX_NB_GVM_REGS; i++)
    pstate->r[i] = ___UND;

  pstate->r[0] = ___UND; /* the kernel replaces this by the break handler */
  pstate->na = 0;
  pstate->pc = ___VECTORREF(___G__23_23exec_2dvector.value,___FIX(0));
}

/*---------------------------------------------------------------------------*/

/* Dispatcher */

void gambit_run()
{
  ___processor ps = pstate;
  ___WORD pc = ps->pc;

  while (___TYP(pc)==___tPROCEDURE)
    {
      ___WORD (*proc)() = ((___WORD (**)())(pc-___tPROCEDURE))[1];
      pc = proc(ps);
      ps->pc = pc;
    }

  if (pc != ___UND)
    if (pc == ___BRK)
      ps->pc = ps->temp1;
    else
      fprintf( stderr,
               "*** ERROR -- Jump to illegal PC (pc=0x%016lx)\n",
               pc );
}

int gambit_alive()
{
  return (___TYP(pstate->pc)==___tPROCEDURE);
}

/*---------------------------------------------------------------------------*/
